import plistlib
from enum import Enum
from pathlib import Path
from typing import Any, Mapping, Optional
from uuid import uuid4

from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.hazmat.primitives.serialization.pkcs7 import PKCS7SignatureBuilder

from pymobiledevice3.exceptions import CloudConfigurationAlreadyPresentError, ProfileError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.lockdown_service import LockdownService

ERROR_CLOUD_CONFIGURATION_ALREADY_PRESENT = 14002
GLOBAL_HTTP_PROXY_UUID = '86a52338-52f7-4c09-b005-52baf3dc4882'
GLOBAL_RESTRICTIONS_UUID = 'e22a0a66-08a8-43f5-9bbc-5279af35bb2b'


class Purpose(Enum):
    PostSetupInstallation = 'PostSetupInstallation'


class MobileConfigService(LockdownService):
    SERVICE_NAME = 'com.apple.mobile.MCInstall'
    RSD_SERVICE_NAME = 'com.apple.mobile.MCInstall.shim.remote'

    def __init__(self, lockdown: LockdownServiceProvider) -> None:
        if isinstance(lockdown, LockdownClient):
            super().__init__(lockdown, self.SERVICE_NAME)
        else:
            super().__init__(lockdown, self.RSD_SERVICE_NAME)

    def hello(self) -> None:
        self._send_recv({'RequestType': 'HelloHostIdentifier'})

    def flush(self) -> None:
        self._send_recv({'RequestType': 'Flush'})

    def escalate(self, keybag_file: Path) -> None:
        """
        Authenticate with the device.

        :param keybag_file: Certificate file in PEM format, containing certificate and private key.
        :return: None
        """
        with open(keybag_file, 'rb') as keybag_file:
            keybag_file = keybag_file.read()
        private_key = serialization.load_pem_private_key(keybag_file, password=None)
        cer = x509.load_pem_x509_certificate(keybag_file)
        public_key = cer.public_bytes(Encoding.DER)
        escalate_response = self._send_recv({
            'RequestType': 'Escalate',
            'SupervisorCertificate': public_key
        })
        signed_challenge = PKCS7SignatureBuilder().set_data(escalate_response['Challenge']).add_signer(
            cer, private_key, hashes.SHA256()).sign(Encoding.DER, [])
        self._send_recv({'RequestType': 'EscalateResponse', 'SignedRequest': signed_challenge})
        self._send_recv({'RequestType': 'ProceedWithKeybagMigration'})

    def get_stored_profile(self, purpose: Purpose = Purpose.PostSetupInstallation) -> Mapping:
        return self._send_recv({'RequestType': 'GetStoredProfile', 'Purpose': purpose.value})

    def store_profile(self, profile_data: bytes, purpose: Purpose = Purpose.PostSetupInstallation) -> None:
        self._send_recv({'RequestType': 'StoreProfile', 'ProfileData': profile_data, 'Purpose': purpose.value})

    def get_cloud_configuration(self) -> Mapping:
        return self._send_recv({'RequestType': 'GetCloudConfiguration'}).get('CloudConfiguration')

    def set_cloud_configuration(self, cloud_configuration: Mapping) -> None:
        self._send_recv({'RequestType': 'SetCloudConfiguration', 'CloudConfiguration': cloud_configuration})

    def establish_provisional_enrollment(self, nonce: bytes) -> None:
        self._send_recv({'RequestType': 'EstablishProvisionalEnrollment', 'Nonce': nonce})

    def set_wifi_power_state(self, state: bool) -> None:
        self._send_recv({'RequestType': 'SetWiFiPowerState', 'PowerState': state})

    def erase_device(self, preserve_data_plan: bool, disallow_proximity_setup: bool) -> None:
        try:
            self._send_recv({'RequestType': 'EraseDevice', 'PreserveDataPlan': preserve_data_plan,
                             'DisallowProximitySetup': disallow_proximity_setup})
        except ConnectionAbortedError:
            pass

    def get_profile_list(self) -> Mapping:
        return self._send_recv({'RequestType': 'GetProfileList'})

    def install_profile(self, payload: bytes) -> None:
        self._send_recv({'RequestType': 'InstallProfile', 'Payload': payload})

    def install_profile_silent(self, keybag_file: Path, profile: bytes) -> None:
        self.escalate(keybag_file)
        self._send_recv({'RequestType': 'InstallProfileSilent', 'Payload': profile})

    def remove_profile(self, identifier: str) -> None:
        profiles = self.get_profile_list()
        if not profiles:
            return
        if identifier not in profiles['ProfileMetadata']:
            self.logger.info(f'Trying to remove not installed profile: {identifier}')
            return
        meta = profiles['ProfileMetadata'][identifier]
        data = plistlib.dumps({'PayloadType': 'Configuration',
                               'PayloadIdentifier': identifier,
                               'PayloadUUID': meta['PayloadUUID'],
                               'PayloadVersion': meta['PayloadVersion']
                               })
        self._send_recv({'RequestType': 'RemoveProfile', 'ProfileIdentifier': data})

    def _send_recv(self, request: Mapping) -> Mapping:
        response = self.service.send_recv_plist(request)
        if response.get('Status', None) != 'Acknowledged':
            error_chain = response.get('ErrorChain')
            if error_chain is not None:
                error_code = error_chain[0]['ErrorCode']
                if error_code == ERROR_CLOUD_CONFIGURATION_ALREADY_PRESENT:
                    raise CloudConfigurationAlreadyPresentError()
            raise ProfileError(f'invalid response {response}')
        return response

    def install_wifi_profile(self, encryption_type: str, ssid: str, password: str,
                             auto_join: bool = True, captive_bypass: bool = False,
                             disable_association_mac_randomization: bool = False, hidden_network: bool = False,
                             is_hotspot: bool = False, keybag_file: Optional[Path] = None) -> None:
        payload_uuid = str(uuid4())
        self.install_managed_profile(f'WiFi Profile For {ssid}', {
            'AutoJoin': auto_join,
            'CaptiveBypass': captive_bypass,
            'DisableAssociationMACRandomization': disable_association_mac_randomization,
            'EncryptionType': encryption_type,
            'HIDDEN_NETWORK': hidden_network,
            'IsHotspot': is_hotspot,
            'Password': password,
            'PayloadDescription': 'Configures Wi-Fi settings',
            'PayloadDisplayName': 'Wi-Fi',
            'PayloadIdentifier': f'com.apple.wifi.managed.{payload_uuid}',
            'PayloadType': 'com.apple.wifi.managed',
            'PayloadUUID': payload_uuid,
            'PayloadVersion': 1,
            'ProxyType': 'None',
            'SSID_STR': ssid
        }, keybag_file=keybag_file)

    def install_http_proxy(self, server: str, server_port: int, keybag_file: Optional[Path] = None) -> None:
        payload_uuid = str(uuid4())
        self.install_managed_profile(f'HTTP Proxy for {server}:{server_port}', {
            'PayloadDescription': 'Global HTTP Proxy',
            'PayloadDisplayName': 'Global HTTP Proxy',
            'PayloadIdentifier': f'com.apple.proxy.http.global.{payload_uuid}',
            'PayloadType': 'com.apple.proxy.http.global',
            'PayloadUUID': payload_uuid,
            'PayloadVersion': 1,
            'ProxyCaptiveLoginAllowed': False,
            'ProxyServer': server,
            'ProxyServerPort': server_port,
            'ProxyType': 'Manual'
        }, payload_uuid=GLOBAL_HTTP_PROXY_UUID, keybag_file=keybag_file)

    def remove_http_proxy(self) -> None:
        self.remove_profile(GLOBAL_HTTP_PROXY_UUID)

    def supervise(self, organization: str, keybag_file: Path) -> None:
        cer = x509.load_pem_x509_certificate(keybag_file.read_bytes())
        public_key = cer.public_bytes(Encoding.DER)
        self.set_cloud_configuration({
            'AllowPairing': True,
            'CloudConfigurationUIComplete': True,
            'ConfigurationSource': 2,
            'ConfigurationWasApplied': True,
            'IsMDMUnremovable': False,
            'IsMandatory': True,
            'IsMultiUser': False,
            'IsSupervised': True,
            'OrganizationMagic': str(uuid4()),
            'OrganizationName': organization,
            'PostSetupProfileWasInstalled': True,
            'SkipSetup': [
                'Location',
                'Restore',
                'SIMSetup',
                'Android',
                'AppleID',
                'IntendedUser',
                'TOS',
                'Siri',
                'ScreenTime',
                'Diagnostics',
                'SoftwareUpdate',
                'Passcode',
                'Biometric',
                'Payment',
                'Zoom',
                'DisplayTone',
                'MessagingActivationUsingPhoneNumber',
                'HomeButtonSensitivity',
                'CloudStorage',
                'ScreenSaver',
                'TapToSetup',
                'Keyboard',
                'PreferredLanguage',
                'SpokenLanguage',
                'WatchMigration',
                'OnBoarding',
                'TVProviderSignIn',
                'TVHomeScreenSync',
                'Privacy',
                'TVRoom',
                'iMessageAndFaceTime',
                'AppStore',
                'Safety',
                'Multitasking',
                'ActionButton',
                'TermsOfAddress',
                'AccessibilityAppearance',
                'Welcome',
                'Appearance',
                'RestoreCompleted',
                'UpdateCompleted',
                'WiFi',
                'Display',
                'Tone',
                'LanguageAndLocale',
                'TouchID',
                'TrueToneDisplay',
                'FileVault',
                'iCloudStorage',
                'iCloudDiagnostics',
                'Registration',
                'DeviceToDeviceMigration',
                'UnlockWithWatch',
                'Accessibility',
                'All',
                'ExpressLanguage',
                'Language',
                'N/A',
                'Region',
                'Avatar',
                'DeviceProtection',
                'Key',
                'LockdownMode',
                'Wallpaper',
                'PrivacySubtitle',
                'SecuritySubtitle',
                'DataSubtitle',
                'AppleIDSubtitle',
                'AppearanceSubtitle',
                'PreferredLang',
                'OnboardingSubtitle',
                'AppleTVSubtitle',
                'Intelligence'
            ],
            'SupervisorHostCertificates': [
                public_key
            ]
        })

    def install_managed_profile(
            self, display_name: str, payload_content: Mapping[str, Any], payload_uuid: str = str(uuid4()),
            keybag_file: Optional[Path] = None) -> None:
        profile_data = plistlib.dumps({
            'PayloadContent': [
                payload_content
            ],
            'PayloadDisplayName': display_name,
            'PayloadIdentifier': payload_uuid,
            'PayloadRemovalDisallowed': False,
            'PayloadType': 'Configuration',
            'PayloadUUID': payload_uuid,
            'PayloadVersion': 1
        })
        if keybag_file is not None:
            self.install_profile_silent(keybag_file, profile_data)
        else:
            self.install_profile(profile_data)

    def install_restrictions_profile(
            self, enforced_software_update_delay: int = 0, payload_uuid: str = GLOBAL_RESTRICTIONS_UUID,
            keybag_file: Optional[Path] = None) -> None:
        self.install_managed_profile('Restrictions', {
            'PayloadDescription': 'Configures restrictions',
            'PayloadDisplayName': 'Restrictions',
            'PayloadIdentifier': f'com.apple.applicationaccess.{payload_uuid}',
            'PayloadType': 'com.apple.applicationaccess',
            'PayloadUUID': payload_uuid,
            'PayloadVersion': 1,
            'allowActivityContinuation': True,
            'allowAddingGameCenterFriends': True,
            'allowAirPlayIncomingRequests': True,
            'allowAirPrint': True,
            'allowAirPrintCredentialsStorage': True,
            'allowAirPrintiBeaconDiscovery': True,
            'allowAppCellularDataModification': True,
            'allowAppClips': True,
            'allowAppInstallation': True,
            'allowAppRemoval': True,
            'allowApplePersonalizedAdvertising': True,
            'allowAssistant': True,
            'allowAssistantWhileLocked': True,
            'allowAutoCorrection': True,
            'allowAutoUnlock': True,
            'allowAutomaticAppDownloads': True,
            'allowBluetoothModification': True,
            'allowBookstore': True,
            'allowBookstoreErotica': True,
            'allowCamera': True,
            'allowCellularPlanModification': True,
            'allowChat': True,
            'allowCloudBackup': True,
            'allowCloudDocumentSync': True,
            'allowCloudPhotoLibrary': True,
            'allowContinuousPathKeyboard': True,
            'allowDefinitionLookup': True,
            'allowDeviceNameModification': True,
            'allowDeviceSleep': True,
            'allowDictation': True,
            'allowESIMModification': True,
            'allowEnablingRestrictions': True,
            'allowEnterpriseAppTrust': True,
            'allowEnterpriseBookBackup': True,
            'allowEnterpriseBookMetadataSync': True,
            'allowEraseContentAndSettings': True,
            'allowExplicitContent': True,
            'allowFilesNetworkDriveAccess': True,
            'allowFilesUSBDriveAccess': True,
            'allowFindMyDevice': True,
            'allowFindMyFriends': True,
            'allowFingerprintForUnlock': True,
            'allowFingerprintModification': True,
            'allowGameCenter': True,
            'allowGlobalBackgroundFetchWhenRoaming': True,
            'allowInAppPurchases': True,
            'allowKeyboardShortcuts': True,
            'allowManagedAppsCloudSync': True,
            'allowMultiplayerGaming': True,
            'allowMusicService': True,
            'allowNews': True,
            'allowNotificationsModification': True,
            'allowOpenFromManagedToUnmanaged': True,
            'allowOpenFromUnmanagedToManaged': True,
            'allowPairedWatch': True,
            'allowPassbookWhileLocked': True,
            'allowPasscodeModification': True,
            'allowPasswordAutoFill': True,
            'allowPasswordProximityRequests': True,
            'allowPasswordSharing': True,
            'allowPersonalHotspotModification': True,
            'allowPhotoStream': True,
            'allowPredictiveKeyboard': True,
            'allowProximitySetupToNewDevice': True,
            'allowRadioService': True,
            'allowRemoteAppPairing': True,
            'allowRemoteScreenObservation': True,
            'allowSafari': True,
            'allowScreenShot': True,
            'allowSharedStream': True,
            'allowSpellCheck': True,
            'allowSpotlightInternetResults': True,
            'allowSystemAppRemoval': True,
            'allowUIAppInstallation': True,
            'allowUIConfigurationProfileInstallation': True,
            'allowUSBRestrictedMode': True,
            'allowUnpairedExternalBootToRecovery': False,
            'allowUntrustedTLSPrompt': True,
            'allowVPNCreation': True,
            'allowVideoConferencing': True,
            'allowVoiceDialing': True,
            'allowWallpaperModification': True,
            'allowiTunes': True,
            'enforcedSoftwareUpdateDelay': enforced_software_update_delay,
            'forceAirDropUnmanaged': False,
            'forceAirPrintTrustedTLSRequirement': False,
            'forceAssistantProfanityFilter': False,
            'forceAuthenticationBeforeAutoFill': False,
            'forceAutomaticDateAndTime': False,
            'forceClassroomAutomaticallyJoinClasses': False,
            'forceClassroomRequestPermissionToLeaveClasses': False,
            'forceClassroomUnpromptedAppAndDeviceLock': False,
            'forceClassroomUnpromptedScreenObservation': False,
            'forceDelayedSoftwareUpdates': True,
            'forceEncryptedBackup': False,
            'forceITunesStorePasswordEntry': False,
            'forceLimitAdTracking': False,
            'forceWatchWristDetection': False,
            'forceWiFiPowerOn': False,
            'forceWiFiWhitelisting': False,
            'ratingApps': 1000,
            'ratingMovies': 1000,
            'ratingRegion': 'us',
            'ratingTVShows': 1000,
            'safariAcceptCookies': 2.0,
            'safariAllowAutoFill': True,
            'safariAllowJavaScript': True,
            'safariAllowPopups': True,
            'safariForceFraudWarning': False
        }, payload_uuid=payload_uuid, keybag_file=keybag_file)



from pymobiledevice3.services.web_protocol.automation_session import By
from tests.services.test_web_protocol.common import LINK_HTML


def test_tag_name(webdriver):
    webdriver.execute_script(f'''document.getElementsByTagName('body')[0].innerHTML = '{LINK_HTML}'; ''')
    element = webdriver.find_element(By.ID, 'id_of_link')
    assert 'a' == element.tag_name



from enum import IntEnum
from typing import List, Mapping, Optional

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.lockdown_service import LockdownService


class InterfaceOrientation(IntEnum):
    PORTRAIT = 1
    PORTRAIT_UPSIDE_DOWN = 2
    LANDSCAPE = 3
    LANDSCAPE_HOME_TO_LEFT = 4


class SpringBoardServicesService(LockdownService):
    RSD_SERVICE_NAME = 'com.apple.springboardservices.shim.remote'
    SERVICE_NAME = 'com.apple.springboardservices'

    def __init__(self, lockdown: LockdownServiceProvider) -> None:
        if isinstance(lockdown, LockdownClient):
            super().__init__(lockdown, self.SERVICE_NAME)
        else:
            super().__init__(lockdown, self.RSD_SERVICE_NAME)

    def get_icon_state(self, format_version: str = '2') -> List:
        cmd = {'command': 'getIconState'}
        if format_version:
            cmd['formatVersion'] = format_version
        return self.service.send_recv_plist(cmd)

    def set_icon_state(self, newstate: Optional[List] = None) -> None:
        if newstate is None:
            newstate = {}
        self.service.send_plist({'command': 'setIconState', 'iconState': newstate})
        self.service.recv_prefixed()

    def get_icon_pngdata(self, bundle_id: str) -> bytes:
        return self.service.send_recv_plist({'command': 'getIconPNGData',
                                             'bundleId': bundle_id}).get('pngData')

    def get_interface_orientation(self) -> InterfaceOrientation:
        res = self.service.send_recv_plist({'command': 'getInterfaceOrientation'})
        return InterfaceOrientation(res.get('interfaceOrientation'))

    def get_wallpaper_pngdata(self) -> bytes:
        return self.service.send_recv_plist({'command': 'getHomeScreenWallpaperPNGData'}).get('pngData')

    def get_homescreen_icon_metrics(self) -> Mapping[str, float]:
        return self.service.send_recv_plist({'command': 'getHomeScreenIconMetrics'})

    def get_wallpaper_info(self, wallpaper_name: str) -> Mapping:
        return self.service.send_recv_plist({'command': 'getWallpaperInfo', 'wallpaperName': wallpaper_name})

    def reload_icon_state(self) -> None:
        self.set_icon_state(self.get_icon_state())

    def get_wallpaper_preview_image(self, wallpaper_name: str) -> bytes:
        return self.service.send_recv_plist({
            'command': 'getWallpaperPreviewImage', 'wallpaperName': wallpaper_name})['pngData']



import logging
import posixpath
import re
import time
from typing import Callable, Generator, List, Optional

from pycrashreport.crash_report import get_crash_report_from_buf
from xonsh.built_ins import XSH
from xonsh.cli_utils import Annotated, Arg

from pymobiledevice3.exceptions import AfcException, NotificationTimeoutError, SysdiagnoseTimeoutError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.afc import AfcService, AfcShell, path_completer
from pymobiledevice3.services.notification_proxy import NotificationProxyService
from pymobiledevice3.services.os_trace import OsTraceService

SYSDIAGNOSE_PROCESS_NAMES = ('sysdiagnose', 'sysdiagnosed')
SYSDIAGNOSE_DIR = 'DiagnosticLogs/sysdiagnose'
SYSDIAGNOSE_IN_PROGRESS_MAX_TTL_SECS = 600

# on iOS17, we need to wait for a moment before trying to fetch the sysdiagnose archive
IOS17_SYSDIAGNOSE_DELAY = 3


class CrashReportsManager:
    COPY_MOBILE_NAME = 'com.apple.crashreportcopymobile'
    RSD_COPY_MOBILE_NAME = 'com.apple.crashreportcopymobile.shim.remote'

    CRASH_MOVER_NAME = 'com.apple.crashreportmover'
    RSD_CRASH_MOVER_NAME = 'com.apple.crashreportmover.shim.remote'

    APPSTORED_PATH = '/com.apple.appstored'
    IN_PROGRESS_SYSDIAGNOSE_EXTENSIONS = ['.tmp', '.tar.gz']

    def __init__(self, lockdown: LockdownServiceProvider):
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown

        if isinstance(lockdown, LockdownClient):
            self.copy_mobile_service_name = self.COPY_MOBILE_NAME
            self.crash_mover_service_name = self.CRASH_MOVER_NAME
        else:
            self.copy_mobile_service_name = self.RSD_COPY_MOBILE_NAME
            self.crash_mover_service_name = self.RSD_CRASH_MOVER_NAME

        self.afc = AfcService(lockdown, service_name=self.copy_mobile_service_name)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self) -> None:
        self.afc.close()

    def clear(self) -> None:
        """
        Clear all crash reports.
        """
        undeleted_items = []
        for filename in self.ls('/'):
            undeleted_items.extend(self.afc.rm(filename, force=True))

        for item in undeleted_items:
            # special case of file that sometimes created automatically right after delete,
            # and then we can't delete the folder because it's not empty
            if item != self.APPSTORED_PATH:
                raise AfcException(f'failed to clear crash reports directory, undeleted items: {undeleted_items}', None)

    def ls(self, path: str = '/', depth: int = 1) -> List[str]:
        """
        List file and folder in the crash report's directory.
        :param path: Path to list, relative to the crash report's directory.
        :param depth: Listing depth, -1 to list infinite.
        :return: List of files listed.
        """
        return list(self.afc.dirlist(path, depth))[1:]  # skip the root path '/'

    def pull(self, out: str, entry: str = '/', erase: bool = False, match: Optional[str] = None) -> None:
        """
        Pull crash reports from the device.
        :param out: Directory to pull crash reports to.
        :param entry: File or Folder to pull.
        :param erase: Whether to erase the original file from the CrashReports directory.
        :param match: Regex to match against file and directory names to pull.
        """

        def log(src: str, dst: str) -> None:
            self.logger.info(f'{src} --> {dst}')
            if erase:
                if not self.afc.isdir(src):
                    self.afc.rm_single(src, force=True)

        match = None if match is None else re.compile(match)
        self.afc.pull(entry, out, match, callback=log)

    def flush(self) -> None:
        """ Trigger com.apple.crashreportmover to flush all products into CrashReports directory """
        ack = b'ping\x00'
        assert ack == self.lockdown.start_lockdown_service(self.crash_mover_service_name).recvall(len(ack))

    def watch(self, name: str = None, raw: bool = False) -> Generator[str, None, None]:
        """
        Monitor creation of new crash reports for a given process name.

        Return value can either be the raw crash string, or parsed result containing a more human-friendly
        representation for the crash.
        """
        for syslog_entry in OsTraceService(lockdown=self.lockdown).syslog():
            if (posixpath.basename(syslog_entry.filename) != 'osanalyticshelper') or \
                    (posixpath.basename(syslog_entry.image_name) != 'OSAnalytics') or \
                    not syslog_entry.message.startswith('Saved type '):
                # skip non-ips creation syslog lines
                continue

            filename = posixpath.basename(syslog_entry.message.split()[-1])
            self.logger.debug(f'crash report: {filename}')

            if posixpath.splitext(filename)[-1] not in ('.ips', '.panic'):
                continue

            crash_report_raw = self.afc.get_file_contents(filename).decode()
            crash_report = get_crash_report_from_buf(crash_report_raw, filename=filename)

            if name is None or crash_report.name == name:
                if raw:
                    yield crash_report_raw
                else:
                    yield crash_report

    def get_new_sysdiagnose(self, out: str, erase: bool = True, *, timeout: Optional[float] = None,
                            callback: Optional[Callable[[float], None]] = None) -> None:
        """
        Monitor the creation of a newly created sysdiagnose archive and pull it
        :param out: filename
        :param erase: remove after pulling
        :keyword timeout: Maximum time in seconds to wait for the completion of sysdiagnose archive
            If None (default), waits indefinitely
        :keyword callback: optional callback function (form: func(float)) that accepts the elapsed time so far
        """
        start_time = time.monotonic()
        end_time = None
        if timeout is not None:
            end_time = start_time + timeout
        sysdiagnose_filename = self._get_new_sysdiagnose_filename(end_time)

        if callback is not None:
            callback(time.monotonic() - start_time)

        self.logger.info('sysdiagnose tarball creation has been started')
        self._wait_for_sysdiagnose_to_finish(timeout)

        if callback is not None:
            callback(time.monotonic() - start_time)

        self.pull(out, entry=sysdiagnose_filename, erase=erase)

        if callback is not None:
            callback(time.monotonic() - start_time)

    def _wait_for_sysdiagnose_to_finish(self, end_time: Optional[float] = None) -> None:
        with NotificationProxyService(self.lockdown, timeout=end_time) as service:
            stop_notification = 'com.apple.sysdiagnose.sysdiagnoseStopped'
            service.notify_register_dispatch(stop_notification)
            try:
                for event in service.receive_notification():
                    if event['Name'] != stop_notification:
                        continue
                    self.logger.debug(f'Received {event}')
                    time.sleep(IOS17_SYSDIAGNOSE_DELAY)
                    break
            except NotificationTimeoutError as e:
                raise SysdiagnoseTimeoutError('Timeout waiting for sysdiagnose completion') from e

    def _get_new_sysdiagnose_filename(self, end_time: Optional[float] = None) -> str:
        sysdiagnose_filename = None
        excluded_temp_files = []

        while sysdiagnose_filename is None:
            try:
                for filename in self.afc.listdir(SYSDIAGNOSE_DIR):
                    # search for an IN_PROGRESS archive
                    if filename not in excluded_temp_files and 'IN_PROGRESS_' in filename:
                        for ext in self.IN_PROGRESS_SYSDIAGNOSE_EXTENSIONS:
                            if filename.endswith(ext):
                                delta = self.lockdown.date - \
                                        self.afc.stat(posixpath.join(SYSDIAGNOSE_DIR, filename))['st_mtime']
                                # Ignores IN_PROGRESS sysdiagnose files older than the defined time to live
                                if delta.total_seconds() < SYSDIAGNOSE_IN_PROGRESS_MAX_TTL_SECS:
                                    self.logger.debug(f'Detected in progress sysdiagnose {filename}')
                                    sysdiagnose_filename = filename.rsplit(ext)[0]
                                    sysdiagnose_filename = sysdiagnose_filename.replace('IN_PROGRESS_', '')
                                    sysdiagnose_filename = f'{sysdiagnose_filename}.tar.gz'
                                    return posixpath.join(SYSDIAGNOSE_DIR, sysdiagnose_filename)
                                else:
                                    self.logger.warning(f"Old sysdiagnose temp file ignored {filename}")
                                    excluded_temp_files.append(filename)
            except AfcException:
                pass

            if self._check_timeout(end_time):
                raise SysdiagnoseTimeoutError('Timeout finding in-progress sysdiagnose filename')

    def _check_timeout(self, end_time: Optional[float] = None) -> bool:
        return end_time is not None and time.monotonic() > end_time


class CrashReportsShell(AfcShell):
    @classmethod
    def create(cls, service_provider: LockdownServiceProvider, **kwargs):
        manager = CrashReportsManager(service_provider)
        XSH.ctx['_manager'] = manager
        super(CrashReportsShell, CrashReportsShell).create(service_provider, service=manager.afc)

    def _setup_shell_commands(self):
        super()._setup_shell_commands()
        self._register_arg_parse_alias('parse', self._do_parse)
        self._register_arg_parse_alias('clear', self._do_clear)

    def _do_parse(self, filename: Annotated[str, Arg(completer=path_completer)]) -> None:
        print(get_crash_report_from_buf(self.afc.get_file_contents(filename).decode(), filename=filename))

    def _do_clear(self) -> None:
        XSH.ctx['_manager'].clear()



import logging

from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.service_connection import ServiceConnection


class LockdownService:
    def __init__(self, lockdown: LockdownServiceProvider, service_name: str, is_developer_service=False,
                 service: ServiceConnection = None, include_escrow_bag: bool = False):
        """
        :param lockdown: server provider
        :param service_name: wrapped service name - will attempt
        :param is_developer_service: should DeveloperDiskImage be mounted before
        :param service: an established service connection object. If none, will attempt connecting to service_name
        """

        if service is None:
            start_service = lockdown.start_lockdown_developer_service if is_developer_service else \
                lockdown.start_lockdown_service
            service = start_service(service_name, include_escrow_bag=include_escrow_bag)

        self.service_name = service_name
        self.lockdown = lockdown
        self.service = service
        self.logger = logging.getLogger(self.__module__)

    def __enter__(self):
        return self

    async def __aenter__(self) -> 'LockdownService':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.service.aio_close()

    def close(self) -> None:
        self.service.close()



import logging
import tempfile

import click

from pymobiledevice3 import usbmux
from pymobiledevice3.cli.cli_common import USBMUX_OPTION_HELP, BaseCommand, print_json
from pymobiledevice3.lockdown import create_using_usbmux
from pymobiledevice3.tcp_forwarder import UsbmuxTcpForwarder

logger = logging.getLogger(__name__)


@click.group()
def cli() -> None:
    pass


@cli.group('usbmux')
def usbmux_cli() -> None:
    """ List devices or forward a TCP port """
    pass


@usbmux_cli.command('forward')
@click.option('usbmux_address', '--usbmux', help=USBMUX_OPTION_HELP)
@click.argument('src_port', type=click.IntRange(1, 0xffff))
@click.argument('dst_port', type=click.IntRange(1, 0xffff))
@click.option('--serial', help='device serial number')
@click.option('-d', '--daemonize', is_flag=True)
def usbmux_forward(usbmux_address: str, src_port: int, dst_port: int, serial: str, daemonize: bool):
    """ forward tcp port """
    forwarder = UsbmuxTcpForwarder(serial, dst_port, src_port, usbmux_address=usbmux_address)

    if daemonize:
        try:
            from daemonize import Daemonize
        except ImportError:
            raise NotImplementedError('daemonizing is only supported on unix platforms')

        with tempfile.NamedTemporaryFile('wt') as pid_file:
            daemon = Daemonize(app=f'forwarder {src_port}->{dst_port}', pid=pid_file.name, action=forwarder.start)
            daemon.start()
    else:
        forwarder.start()


@usbmux_cli.command('list', cls=BaseCommand)
@click.option('usbmux_address', '--usbmux', help=USBMUX_OPTION_HELP)
@click.option('-u', '--usb', is_flag=True, help='show only usb devices')
@click.option('-n', '--network', is_flag=True, help='show only network devices')
def usbmux_list(usbmux_address: str, usb: bool, network: bool) -> None:
    """ list connected devices """
    connected_devices = []
    for device in usbmux.list_devices(usbmux_address=usbmux_address):
        udid = device.serial

        if usb and not device.is_usb:
            continue

        if network and not device.is_network:
            continue

        lockdown = create_using_usbmux(udid, autopair=False, connection_type=device.connection_type,
                                       usbmux_address=usbmux_address)
        connected_devices.append(lockdown.short_info)

    print_json(connected_devices)



import logging
import time
from typing import Any, Mapping, Optional

from bpylist2 import archiver
from packaging.version import Version

from pymobiledevice3.exceptions import AppNotInstalledError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.dvt.dvt_secure_socket_proxy import DvtSecureSocketProxyService
from pymobiledevice3.services.dvt.dvt_testmanaged_proxy import DvtTestmanagedProxyService
from pymobiledevice3.services.dvt.instruments.process_control import ProcessControl
from pymobiledevice3.services.house_arrest import HouseArrestService
from pymobiledevice3.services.installation_proxy import InstallationProxyService
from pymobiledevice3.services.remote_server import NSURL, NSUUID, Channel, ChannelFragmenter, MessageAux, \
    XCTestConfiguration, dtx_message_header_struct, dtx_message_payload_header_struct

logger = logging.getLogger(__name__)


class XCUITestService:
    IDENTIFIER = "dtxproxy:XCTestManager_IDEInterface:XCTestManager_DaemonConnectionInterface"
    XCODE_VERSION = 36  # not important

    def __init__(self, service_provider: LockdownServiceProvider):
        self.service_provider = service_provider
        self.pctl = self.init_process_control()
        self.product_major_version = Version(service_provider.product_version).major

    def run(
        self,
        bundle_id: str,
        test_runner_env: Optional[dict] = None,
        test_runner_args: Optional[list] = None,
    ):
        # Test OK with
        # - iPhone SE (iPhone8,4) 15.8
        #
        # Test Failed with
        # - iPhone 12 Pro (iPhone13,3) 17.2
        #
        # TODO: it seems the protocol changed when iOS>=17
        session_identifier = NSUUID.uuid4()
        app_info = get_app_info(self.service_provider, bundle_id)

        xctest_configuration = generate_xctestconfiguration(
            app_info, session_identifier, bundle_id, test_runner_env, test_runner_args
        )
        xctest_path = f"/tmp/{str(session_identifier).upper()}.xctestconfiguration"  # yapf: disable

        self.setup_xcuitest(bundle_id, xctest_path, xctest_configuration)
        dvt1, chan1, dvt2, chan2 = self.init_ide_channels(session_identifier)

        pid = self.launch_test_app(
            app_info, bundle_id, xctest_path, test_runner_env, test_runner_args
        )
        logger.info("Runner started with pid:%d, waiting for testBundleReady", pid)

        time.sleep(1)
        self.authorize_test_process_id(chan1, pid)
        self.start_executing_test_plan_with_protocol_version(dvt2, self.XCODE_VERSION)

        # TODO: boradcast message is not handled
        # TODO: RemoteServer.receive_message is not thread safe and will block if no message received
        try:
            self.dispatch(dvt2, chan2)
        except KeyboardInterrupt:
            logger.info("Signal Interrupt catched")
        finally:
            logger.info("Killing UITest with pid %d ...", pid)
            self.pctl.kill(pid)
            dvt1.close()
            dvt2.close()

    def dispatch(self, dvt: DvtTestmanagedProxyService, chan: Channel):
        while True:
            self.dispatch_proxy(dvt, chan)

    def dispatch_proxy(self, dvt: DvtTestmanagedProxyService, chan: Channel):
        # Ref code:
        # https://github.com/danielpaulus/go-ios/blob/a49a3582ef4438fee794912c167d2cccf45d8efa/ios/testmanagerd/xcuitestrunner.go#L182
        # https://github.com/alibaba/tidevice/blob/main/tidevice/_device.py#L1117

        key, value = dvt.recv_plist(chan)
        value = value and value[0].value.strip()
        if key == "_XCT_logDebugMessage:":
            logger.debug("logDebugMessage: %s", value)
        elif key == "_XCT_testRunnerReadyWithCapabilities:":
            logger.info("testRunnerReadyWithCapabilities: %s", value)
            self.send_response_capabilities(dvt, chan, dvt.cur_message)
        else:
            # There are still unhandled messages
            # - _XCT_testBundleReadyWithProtocolVersion:minimumVersion:
            # - _XCT_didFinishExecutingTestPlan
            logger.info("unhandled %s %r", key, value)

    def send_response_capabilities(
        self, dvt: DvtTestmanagedProxyService, chan: Channel, cur_message: int
    ):
        pheader = dtx_message_payload_header_struct.build(
            dict(flags=3, auxiliaryLength=0, totalLength=0)
        )
        mheader = dtx_message_header_struct.build(
            dict(
                cb=dtx_message_header_struct.sizeof(),
                fragmentId=0,
                fragmentCount=1,
                length=dtx_message_payload_header_struct.sizeof(),
                identifier=cur_message,
                conversationIndex=1,
                channelCode=chan,
                expectsReply=int(0),
            )
        )
        msg = mheader + pheader
        dvt.service.sendall(msg)

    def init_process_control(self):
        dvt_proxy = DvtSecureSocketProxyService(lockdown=self.service_provider)
        dvt_proxy.perform_handshake()
        return ProcessControl(dvt_proxy)

    def init_ide_channels(self, session_identifier: NSUUID):
        # XcodeIDE require two connections
        dvt1 = DvtTestmanagedProxyService(lockdown=self.service_provider)
        dvt1.perform_handshake()

        logger.info("make channel %s", self.IDENTIFIER)
        chan1 = dvt1.make_channel(self.IDENTIFIER)
        if self.product_major_version >= 11:
            dvt1.send_message(
                chan1,
                "_IDE_initiateControlSessionWithProtocolVersion:",
                MessageAux().append_obj(self.XCODE_VERSION),
            )
            reply = chan1.receive_plist()
            logger.info("conn1 handshake xcode version: %s", reply)

        dvt2 = DvtTestmanagedProxyService(lockdown=self.service_provider)
        dvt2.perform_handshake()
        chan2 = dvt2.make_channel(self.IDENTIFIER)
        dvt2.send_message(
            channel=chan2,
            selector="_IDE_initiateSessionWithIdentifier:forClient:atPath:protocolVersion:",
            args=MessageAux()
            .append_obj(session_identifier)
            .append_obj("not-very-import-part")  # this part is not important
            .append_obj("/Applications/Xcode.app/Contents/Developer/usr/bin/xcodebuild")
            .append_obj(self.XCODE_VERSION),
        )
        reply = chan2.receive_plist()
        logger.info("conn2 handshake xcode version: %s", reply)
        return dvt1, chan1, dvt2, chan2

    def setup_xcuitest(
        self,
        bundle_id: str,
        xctest_path: str,
        xctest_configuration: XCTestConfiguration,
    ):
        """push xctestconfiguration to app VendDocuments"""
        with HouseArrestService(
            lockdown=self.service_provider, bundle_id=bundle_id, documents_only=False
        ) as afc:
            for name in afc.listdir("/tmp"):
                if name.endswith(".xctestconfiguration"):
                    logger.debug("remove /tmp/%s", name)
                    afc.rm("/tmp/" + name)
            afc.set_file_contents(xctest_path, archiver.archive(xctest_configuration))

    def start_executing_test_plan_with_protocol_version(self, dvt: DvtTestmanagedProxyService, protocol_version: int):
        ide_channel = Channel.create(-1, dvt)
        dvt.channel_messages[ide_channel] = ChannelFragmenter()
        dvt.send_message(
            ide_channel,
            "_IDE_startExecutingTestPlanWithProtocolVersion:",
            MessageAux().append_obj(protocol_version),
            expects_reply=False,
        )

    def authorize_test_process_id(self, chan: Channel, pid: int):
        selector = None
        aux = MessageAux()
        if self.product_major_version >= 12:
            selector = "_IDE_authorizeTestSessionWithProcessID:"
            aux.append_obj(pid)
        elif self.product_major_version >= 10:
            selector = "_IDE_initiateControlSessionForTestProcessID:protocolVersion:"
            aux.append_obj(pid)
            aux.append_obj(self.XCODE_VERSION)
        else:
            selector = "_IDE_initiateControlSessionForTestProcessID:"
            aux.append_obj(pid)
        chan.send_message(selector, aux)
        reply = chan.receive_plist()
        if isinstance(reply, bool) and reply is True:
            logger.info("authorizing test session for pid %d successful %r", pid, reply)
        else:
            raise RuntimeError("Failed to authorize test process id: %s" % reply)

    def launch_test_app(
        self,
        app_info: dict,
        bundle_id: str,
        xctest_path: str,
        test_runner_env: Optional[dict] = None,
        test_runner_args: Optional[list] = None,
    ) -> int:
        app_container = app_info["Container"]
        app_path = app_info["Path"]
        exec_name = app_info["CFBundleExecutable"]
        # # logger.info('CFBundleExecutable: %s', exec_name)
        # # CFBundleName always endswith -Runner
        assert exec_name.endswith("-Runner"), (
            "Invalid CFBundleExecutable: %s" % exec_name
        )
        target_name = exec_name[: -len("-Runner")]

        app_env = {
            "CA_ASSERT_MAIN_THREAD_TRANSACTIONS": "0",
            "CA_DEBUG_TRANSACTIONS": "0",
            "DYLD_FRAMEWORK_PATH": app_path + "/Frameworks:",
            "DYLD_LIBRARY_PATH": app_path + "/Frameworks",
            "MTC_CRASH_ON_REPORT": "1",
            "NSUnbufferedIO": "YES",
            "SQLITE_ENABLE_THREAD_ASSERTIONS": "1",
            "WDA_PRODUCT_BUNDLE_IDENTIFIER": "",
            "XCTestBundlePath": f'{app_info["Path"]}/PlugIns/{target_name}.xctest',
            "XCTestConfigurationFilePath": app_container + xctest_path,
            "XCODE_DBG_XPC_EXCLUSIONS": "com.apple.dt.xctestSymbolicator",
            # the following maybe no needed
            # 'MJPEG_SERVER_PORT': '',
            # 'USE_PORT': '',
            # 'LLVM_PROFILE_FILE': app_container + '/tmp/%p.profraw', # %p means pid
        }
        if test_runner_env:
            app_env.update(test_runner_env)

        if self.product_major_version >= 11:
            app_env[
                "DYLD_INSERT_LIBRARIES"
            ] = "/Developer/usr/lib/libMainThreadChecker.dylib"
            app_env["OS_ACTIVITY_DT_MODE"] = "YES"

        app_args = [
            "-NSTreatUnknownArgumentsAsOpen",
            "NO",
            "-ApplePersistenceIgnoreState",
            "YES",
        ]
        app_args.extend(test_runner_args or [])
        app_options = {"StartSuspendedKey": False}
        if self.product_major_version >= 12:
            app_options["ActivateSuspended"] = True

        pid = self.pctl.launch(
            bundle_id,
            arguments=app_args,
            environment=app_env,
            extra_options=app_options,
        )
        for message in self.pctl:
            logger.info("ProcessOutput: %s", message)
        return pid


def get_app_info(service_provider: LockdownClient, bundle_id: str) -> Mapping[str, Any]:
    with InstallationProxyService(lockdown=service_provider) as install_service:
        apps = install_service.get_apps(bundle_identifiers=[bundle_id])
        if not apps:
            raise AppNotInstalledError(f"No app with bundle id {bundle_id} found")
        return apps[bundle_id]


def generate_xctestconfiguration(
    app_info: dict,
    session_identifier: NSUUID,
    target_app_bundle_id: str = None,
    target_app_env: Optional[dict] = None,
    target_app_args: Optional[list] = None,
    tests_to_run: Optional[list] = None,
) -> XCTestConfiguration:
    exec_name: str = app_info["CFBundleExecutable"]
    assert exec_name.endswith("-Runner"), "Invalid CFBundleExecutable: %s" % exec_name
    config_name = exec_name[: -len("-Runner")]

    return XCTestConfiguration(
        {
            "testBundleURL": NSURL(
                None, f'file://{app_info["Path"]}/PlugIns/{config_name}.xctest'
            ),
            "sessionIdentifier": session_identifier,
            "targetApplicationBundleID": target_app_bundle_id,
            "targetApplicationEnvironment": target_app_env or {},
            "targetApplicationArguments": target_app_args or [],
            "testsToRun": tests_to_run or set(),
            "testsMustRunOnMainThread": True,
            "reportResultsToIDE": True,
            "reportActivities": True,
            "automationFrameworkPath": "/Developer/Library/PrivateFrameworks/XCTAutomationSupport.framework",
        }
    )



from pymobiledevice3.services.dvt.instruments.screenshot import Screenshot

PNG_HEADER = b'\x89\x50\x4e\x47\x0d\x0a\x1a\x0a'
TIFF_HEADER = b'\x4D\x4D\x00\x2A'


def test_screenshot(dvt):
    """
    Test that taking a screenshot returns a PNG.
    :param pymobiledevice3.lockdown.LockdownClient lockdown: Lockdown client.
    """
    screenshot = Screenshot(dvt).get_screenshot()
    assert screenshot.startswith(PNG_HEADER) or screenshot.startswith(TIFF_HEADER)



import asyncio
import contextlib
import logging
import os
import plistlib
import tempfile
import traceback
from pathlib import Path
from typing import IO, Generator, Optional, Union
from zipfile import ZipFile

import click
import IPython
import requests
from pygments import formatters, highlight, lexers

from pymobiledevice3 import usbmux
from pymobiledevice3.cli.cli_common import print_json, prompt_selection, set_verbosity
from pymobiledevice3.exceptions import ConnectionFailedError, ConnectionFailedToUsbmuxdError, IncorrectModeError
from pymobiledevice3.irecv import IRecv
from pymobiledevice3.lockdown import LockdownClient, create_using_usbmux
from pymobiledevice3.restore.device import Device
from pymobiledevice3.restore.recovery import Behavior, Recovery
from pymobiledevice3.restore.restore import Restore
from pymobiledevice3.services.diagnostics import DiagnosticsService
from pymobiledevice3.utils import file_download

SHELL_USAGE = """
# use `irecv` variable to access Restore mode API
# for example:
print(irecv.getenv('build-version'))
"""

logger = logging.getLogger(__name__)
IPSWME_API = 'https://api.ipsw.me/v4/device/'


class Command(click.Command):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.params[:0] = [
            click.Option(('device', '--ecid'), type=click.INT, callback=self.device),
            click.Option(('verbosity', '-v', '--verbose'), count=True, callback=set_verbosity, expose_value=False),
        ]

    @staticmethod
    def device(ctx, param, value) -> Optional[Union[LockdownClient, IRecv]]:
        if '_PYMOBILEDEVICE3_COMPLETE' in os.environ:
            # prevent lockdown connection establishment when in autocomplete mode
            return

        ecid = value
        logger.debug('searching among connected devices via lockdownd')
        devices = [dev for dev in usbmux.list_devices() if dev.connection_type == 'USB']
        if len(devices) > 1:
            raise click.ClickException('Multiple device detected')
        try:
            for device in devices:
                try:
                    lockdown = create_using_usbmux(serial=device.serial, connection_type='USB')
                except (ConnectionFailedError, IncorrectModeError):
                    continue
                if (ecid is None) or (lockdown.ecid == value):
                    logger.debug('found device')
                    return lockdown
                else:
                    continue
        except ConnectionFailedToUsbmuxdError:
            pass

        logger.debug('waiting for device to be available in Recovery mode')
        return IRecv(ecid=ecid)


@contextlib.contextmanager
def tempzip_download_ctx(url: str) -> Generator[ZipFile, None, None]:
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpzip = Path(tmpdir) / url.split('/')[-1]
        file_download(url, tmpzip)
        yield ZipFile(tmpzip)


@contextlib.contextmanager
def zipfile_ctx(path: str) -> Generator[ZipFile, None, None]:
    yield ZipFile(path)


class IPSWCommand(Command):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.params.extend([click.Option(('ipsw_ctx', '-i', '--ipsw'), required=False,
                                         callback=self.ipsw_ctx, help='local IPSW file'),
                            click.Option(('tss', '--tss'), type=click.File('rb'), callback=self.tss)])

    @staticmethod
    def ipsw_ctx(ctx, param, value) -> Generator[ZipFile, None, None]:
        if value and not value.startswith(('http://', 'https://')):
            return zipfile_ctx(value)

        url = value
        if url is None:
            url = query_ipswme(ctx.params['device'].product_type)
        return tempzip_download_ctx(url)

    @staticmethod
    def tss(ctx, param, value) -> Optional[IO]:
        if value is None:
            return
        return plistlib.load(value)


def query_ipswme(identifier: str) -> str:
    resp = requests.get(IPSWME_API + identifier, headers={'Accept': 'application/json'})
    firmwares = resp.json()['firmwares']
    display_list = [f'{entry["version"]}: {entry["buildid"]}' for entry in firmwares if entry['signed']]
    idx = prompt_selection(display_list, 'Choose version', idx=True)
    return firmwares[idx]['url']


async def restore_update_task(device: Device, ipsw: ZipFile, tss: Optional[IO], erase: bool, ignore_fdr: bool) -> None:
    lockdown = None
    irecv = None
    if isinstance(device, LockdownClient):
        lockdown = device
    elif isinstance(device, IRecv):
        irecv = device
    device = Device(lockdown=lockdown, irecv=irecv)

    behavior = Behavior.Update
    if erase:
        behavior = Behavior.Erase

    try:
        await Restore(ipsw, device, tss=tss, behavior=behavior, ignore_fdr=ignore_fdr).update()
    except Exception:
        # click may "swallow" several exception types so we try to catch them all here
        traceback.print_exc()
        raise


@click.group()
def cli() -> None:
    pass


@cli.group()
def restore() -> None:
    """ Restore an IPSW or access device in recovery mode """
    pass


@restore.command('shell', cls=Command)
def restore_shell(device):
    """ create an IPython shell for interacting with iBoot """
    IPython.embed(
        header=highlight(SHELL_USAGE, lexers.PythonLexer(), formatters.TerminalTrueColorFormatter(style='native')),
        user_ns={
            'irecv': device,
        })


@restore.command('enter', cls=Command)
def restore_enter(device):
    """ enter Recovery mode """
    if isinstance(device, LockdownClient):
        device.enter_recovery()


@restore.command('exit')
def restore_exit():
    """ exit Recovery mode """
    irecv = IRecv()
    irecv.set_autoboot(True)
    irecv.reboot()


@restore.command('restart', cls=Command)
def restore_restart(device):
    """ restarts device """
    if isinstance(device, LockdownClient):
        with DiagnosticsService(device) as diagnostics:
            diagnostics.restart()
    else:
        device.reboot()


@restore.command('tss', cls=IPSWCommand)
@click.argument('out', type=click.File('wb'), required=False)
def restore_tss(device: Device, ipsw_ctx: Generator, out):
    """ query SHSH blobs """
    lockdown = None
    irecv = None
    if isinstance(device, LockdownClient):
        lockdown = device
    elif isinstance(device, IRecv):
        irecv = device

    device = Device(lockdown=lockdown, irecv=irecv)
    with ipsw_ctx as ipsw:
        tss = Recovery(ipsw, device).fetch_tss_record()
    if out:
        plistlib.dump(tss, out)
    print_json(tss)


@restore.command('ramdisk', cls=IPSWCommand)
def restore_ramdisk(device: Device, ipsw_ctx: Generator, tss: IO):
    """
    don't perform an actual restore. just enter the update ramdisk

    ipsw can be either a filename or an url
    """
    lockdown = None
    irecv = None
    if isinstance(device, LockdownClient):
        lockdown = device
    elif isinstance(device, IRecv):
        irecv = device
    device = Device(lockdown=lockdown, irecv=irecv)
    with ipsw_ctx as ipsw:
        Recovery(ipsw, device, tss=tss).boot_ramdisk()


@restore.command('update', cls=IPSWCommand)
@click.option('--erase', is_flag=True, help='use the Erase BuildIdentity (full factory-reset)')
@click.option('--ignore-fdr', is_flag=True, help='only establish an FDR service connection, but don\'t proxy any '
                                                 'traffic')
def restore_update(device: Device, ipsw_ctx: Generator, tss: IO, erase: bool, ignore_fdr: bool) -> None:
    """
    perform an update

    ipsw can be either a filename or an url
    """
    with ipsw_ctx as ipsw:
        asyncio.run(restore_update_task(device, ipsw, tss, erase, ignore_fdr), debug=True)



import asyncio
import sys
from asyncio import IncompleteReadError
from typing import Generator, Mapping, Optional, Tuple

import IPython
import nest_asyncio
from construct import StreamError
from hyperframe.frame import DataFrame, Frame, GoAwayFrame, HeadersFrame, RstStreamFrame, SettingsFrame, \
    WindowUpdateFrame
from pygments import formatters, highlight, lexers
from traitlets.config import Config

from pymobiledevice3.exceptions import StreamClosedError
from pymobiledevice3.remote.xpc_message import XpcFlags, XpcInt64Type, XpcUInt64Type, XpcWrapper, create_xpc_wrapper, \
    decode_xpc_object

# Extracted by sniffing `remoted` traffic via Wireshark
DEFAULT_SETTINGS_MAX_CONCURRENT_STREAMS = 100
DEFAULT_SETTINGS_INITIAL_WINDOW_SIZE = 1048576
DEFAULT_WIN_SIZE_INCR = 983041

FRAME_HEADER_SIZE = 9
HTTP2_MAGIC = b'PRI * HTTP/2.0\r\n\r\nSM\r\n\r\n'

ROOT_CHANNEL = 1
REPLY_CHANNEL = 3

SHELL_USAGE = """
# This shell allows you to communicate directly with every RemoteXPC service.

# For example, you can do the following:
resp = await client.send_receive_request({"Command": "DoSomething"})
"""


class RemoteXPCConnection:
    def __init__(self, address: Tuple[str, int]):
        self._previous_frame_data = b''
        self.address = address
        self.next_message_id: Mapping[int: int] = {ROOT_CHANNEL: 0, REPLY_CHANNEL: 0}
        self.peer_info = None
        self._reader: Optional[asyncio.StreamReader] = None
        self._writer: Optional[asyncio.StreamWriter] = None

    async def __aenter__(self) -> 'RemoteXPCConnection':
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()

    async def connect(self) -> None:
        self._reader, self._writer = await asyncio.open_connection(self.address[0], self.address[1])
        await self._do_handshake()

    async def close(self) -> None:
        if self._writer is None:
            return
        self._writer.close()
        try:
            await self._writer.wait_closed()
        except ConnectionResetError:
            pass
        self._writer = None
        self._reader = None

    async def send_request(self, data: Mapping, wanting_reply: bool = False) -> None:
        xpc_wrapper = create_xpc_wrapper(
            data, message_id=self.next_message_id[ROOT_CHANNEL], wanting_reply=wanting_reply)
        self._writer.write(DataFrame(stream_id=ROOT_CHANNEL, data=xpc_wrapper).serialize())
        await self._writer.drain()

    async def iter_file_chunks(self, total_size: int, file_idx: int = 0) -> Generator[bytes, None, None]:
        stream_id = (file_idx + 1) * 2
        await self._open_channel(stream_id, XpcFlags.FILE_TX_STREAM_RESPONSE)
        size = 0
        while size < total_size:
            frame = await self._receive_next_data_frame()

            if 'END_STREAM' in frame.flags:
                continue

            if frame.stream_id != stream_id:
                xpc_wrapper = XpcWrapper.parse(frame.data)
                if xpc_wrapper.flags.FILE_TX_STREAM_REQUEST:
                    continue

            assert frame.stream_id == stream_id, f'got {frame.stream_id} instead of {stream_id}'
            size += len(frame.data)
            yield frame.data

    async def receive_file(self, total_size: int) -> bytes:
        buf = b''
        async for chunk in self.iter_file_chunks(total_size):
            buf += chunk
        return buf

    async def receive_response(self) -> Mapping:
        while True:
            frame = await self._receive_next_data_frame()
            try:
                xpc_message = XpcWrapper.parse(self._previous_frame_data + frame.data).message
                self._previous_frame_data = b''
            except StreamError:
                self._previous_frame_data += frame.data
                continue
            if xpc_message.payload is None:
                continue
            if xpc_message.payload.obj.data.entries is None:
                continue
            self.next_message_id[frame.stream_id] = xpc_message.message_id + 1
            return decode_xpc_object(xpc_message.payload.obj)

    async def send_receive_request(self, data: Mapping) -> Mapping:
        await self.send_request(data, wanting_reply=True)
        return await self.receive_response()

    def shell(self) -> None:
        nest_asyncio.apply(asyncio.get_running_loop())
        sys.argv = ['a']
        config = Config()
        config.InteractiveShellApp.exec_lines = ['%autoawait asyncio']
        print(highlight(SHELL_USAGE, lexers.PythonLexer(),
                        formatters.TerminalTrueColorFormatter(style='native')))
        IPython.start_ipython(config=config, user_ns={
            'client': self,
            'XpcInt64Type': XpcInt64Type,
            'XpcUInt64Type': XpcUInt64Type,
        })

    async def _do_handshake(self) -> None:
        self._writer.write(HTTP2_MAGIC)
        await self._writer.drain()

        # send h2 headers
        await self._send_frame(SettingsFrame(settings={
            SettingsFrame.MAX_CONCURRENT_STREAMS: DEFAULT_SETTINGS_MAX_CONCURRENT_STREAMS,
            SettingsFrame.INITIAL_WINDOW_SIZE: DEFAULT_SETTINGS_INITIAL_WINDOW_SIZE,
        }))
        await self._send_frame(WindowUpdateFrame(stream_id=0, window_increment=DEFAULT_WIN_SIZE_INCR))
        await self._send_frame(HeadersFrame(stream_id=ROOT_CHANNEL, flags=['END_HEADERS']))

        # send first actual requests
        await self.send_request({})
        await self._send_frame(DataFrame(stream_id=ROOT_CHANNEL,
                                         data=XpcWrapper.build({'size': 0, 'flags': 0x0201, 'payload': None})))
        self.next_message_id[ROOT_CHANNEL] += 1
        await self._open_channel(REPLY_CHANNEL, XpcFlags.INIT_HANDSHAKE)
        self.next_message_id[REPLY_CHANNEL] += 1

        assert isinstance(await self._receive_frame(), SettingsFrame)

        await self._send_frame(SettingsFrame(flags=['ACK']))

    async def _open_channel(self, stream_id: int, flags: int) -> None:
        flags |= XpcFlags.ALWAYS_SET
        await self._send_frame(HeadersFrame(stream_id=stream_id, flags=['END_HEADERS']))
        await self._send_frame(
            DataFrame(stream_id=stream_id, data=XpcWrapper.build({'size': 0, 'flags': flags, 'payload': None})))

    async def _send_frame(self, frame: Frame) -> None:
        self._writer.write(frame.serialize())
        await self._writer.drain()

    async def _receive_next_data_frame(self) -> DataFrame:
        while True:
            frame = await self._receive_frame()

            if isinstance(frame, GoAwayFrame):
                raise StreamClosedError(f'Got {frame}')
            if isinstance(frame, RstStreamFrame):
                raise StreamClosedError(f'Got {frame}')
            if not isinstance(frame, DataFrame):
                continue

            if frame.stream_id % 2 == 0 and frame.body_len > 0:
                await self._send_frame(WindowUpdateFrame(stream_id=0, window_increment=frame.body_len))
                await self._send_frame(WindowUpdateFrame(stream_id=frame.stream_id, window_increment=frame.body_len))

            return frame

    async def _receive_frame(self) -> Frame:
        buf = await self._reader.readexactly(FRAME_HEADER_SIZE)
        frame, additional_size = Frame.parse_frame_header(memoryview(buf))
        frame.parse_body(memoryview(await self._recvall(additional_size)))
        return frame

    async def _recvall(self, size: int) -> bytes:
        data = b''
        while len(data) < size:
            try:
                chunk = await self._reader.readexactly(size - len(data))
            except IncompleteReadError:
                raise ConnectionAbortedError()
            data += chunk
        return data



#!/usr/bin/env python3

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.lockdown_service import LockdownService


class PreboardService(LockdownService):
    RSD_SERVICE_NAME = 'com.apple.preboardservice_v2.shim.remote'
    SERVICE_NAME = 'com.apple.preboardservice_v2'

    def __init__(self, lockdown: LockdownServiceProvider):
        if isinstance(lockdown, LockdownClient):
            super().__init__(lockdown, self.SERVICE_NAME)
        else:
            super().__init__(lockdown, self.RSD_SERVICE_NAME)

    def create_stashbag(self, manifest):
        return self.service.send_recv_plist({'Command': 'CreateStashbag', 'Manifest': manifest})

    def commit(self, manifest):
        return self.service.send_recv_plist({'Command': 'CommitStashbag', 'Manifest': manifest})



import pytest
from packaging.version import Version

from pymobiledevice3.bonjour import browse_mobdev2, browse_remoted, browse_remotepairing


@pytest.mark.asyncio
async def test_mobdev2(lockdown):
    lockdown.enable_wifi_connections = True
    assert len(await browse_mobdev2()) >= 1


@pytest.mark.asyncio
async def test_remoted(lockdown):
    if Version(lockdown.product_version) < Version('16.0'):
        pytest.skip('iOS < 16.0')
    assert len(await browse_remoted()) >= 1


@pytest.mark.asyncio
async def test_remotepairing(lockdown):
    if Version(lockdown.product_version) < Version('17.0'):
        pytest.skip('iOS < 17.0')
    assert len(await browse_remotepairing()) >= 1



from pymobiledevice3.services.dvt.instruments.core_profile_session_tap import CoreProfileSessionTap


def test_stackshot(dvt):
    """
    Test getting stackshot.
    """
    with CoreProfileSessionTap(dvt, CoreProfileSessionTap.get_time_config(dvt)) as tap:
        data = tap.get_stackshot()

    assert 'Darwin Kernel' in data['osversion']
    # Constant kernel task data.
    assert data['task_snapshots'][0]['task_snapshot']['ts_pid'] == 0
    assert data['task_snapshots'][0]['task_snapshot']['ts_p_comm'] == 'kernel_task'



import datetime
import os
import socket
from pathlib import Path
from typing import List, Tuple

import win32security
from ifaddr import get_adapters

from pymobiledevice3.osu.os_utils import DEFAULT_AFTER_IDLE_SEC, DEFAULT_INTERVAL_SEC, OsUtils
from pymobiledevice3.usbmux import MuxConnection


class Win32(OsUtils):
    @property
    def is_admin(self) -> bool:
        """ Check if the current OS user is an Administrator or root.
        See: https://github.com/Preston-Landers/pyuac/blob/master/pyuac/admin.py
        :return: True if the current user is an 'Administrator', otherwise False.
        """
        try:
            admin_sid = win32security.CreateWellKnownSid(win32security.WinBuiltinAdministratorsSid, None)
            return win32security.CheckTokenMembership(None, admin_sid)
        except Exception:
            return False

    @property
    def usbmux_address(self) -> Tuple[str, int]:
        return MuxConnection.ITUNES_HOST, socket.AF_INET

    @property
    def bonjour_timeout(self) -> int:
        return 2

    @property
    def loopback_header(self) -> bytes:
        return b'\x00\x00\x86\xdd'

    @property
    def access_denied_error(self) -> str:
        return 'This command requires admin privileges. Consider retrying with "run-as administrator".'

    @property
    def pair_record_path(self) -> Path:
        return Path(os.environ.get('ALLUSERSPROFILE', ''), 'Apple', 'Lockdown')

    def get_ipv6_ips(self) -> List[str]:
        return [f'{adapter.ips[0].ip[0]}%{adapter.ips[0].ip[2]}' for adapter in get_adapters() if
                adapter.ips[0].is_IPv6]

    def set_keepalive(self, sock: socket.socket, after_idle_sec: int = DEFAULT_AFTER_IDLE_SEC,
                      interval_sec: int = DEFAULT_INTERVAL_SEC, **kwargs) -> None:
        sock.ioctl(socket.SIO_KEEPALIVE_VALS, (1, after_idle_sec * 1000, interval_sec * 1000))

    def parse_timestamp(self, time_stamp) -> datetime:
        return datetime.datetime.fromtimestamp(time_stamp / 1000)

    def chown_to_non_sudo_if_needed(self, path: Path) -> None:
        return

    def wait_return(self):
        input('Press ENTER to exit>')



import click


@click.group()
def cli() -> None:
    pass


@cli.command()
def version() -> None:
    """ Query pymobiledevice3 version """
    try:
        from pymobiledevice3._version import __version__
        print(__version__)
    except ImportError:
        print('version could not be determined. please first install/build the package')



import click

from pymobiledevice3.cli.cli_common import Command
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.mobile_activation import MobileActivationService


@click.group()
def cli() -> None:
    pass


@cli.group()
def activation() -> None:
    """ Perform iCloud activation/deactivation or query the current state """
    pass


@activation.command(cls=Command)
def state(service_provider: LockdownClient):
    """ Get current activation state """
    print(MobileActivationService(service_provider).state)


@activation.command(cls=Command)
@click.option('--now', is_flag=True, help='when --offline is used, dont wait for next nonce cycle')
def activate(service_provider: LockdownClient, now):
    """ Activate device """
    activation_service = MobileActivationService(service_provider)
    if not now:
        activation_service.wait_for_activation_session()
    activation_service.activate()


@activation.command(cls=Command)
def deactivate(service_provider: LockdownClient):
    """ Deactivate device """
    MobileActivationService(service_provider).deactivate()



import dataclasses
import uuid
from pathlib import Path
from typing import List

from tqdm import tqdm

from pymobiledevice3.remote.remote_service import RemoteService
from pymobiledevice3.remote.remote_service_discovery import RemoteServiceDiscoveryService


@dataclasses.dataclass
class DSCFile:
    file_path: str
    file_size: int


class RemoteFetchSymbolsService(RemoteService):
    SERVICE_NAME = 'com.apple.dt.remoteFetchSymbols'

    def __init__(self, rsd: RemoteServiceDiscoveryService):
        super().__init__(rsd, self.SERVICE_NAME)

    async def get_dsc_file_list(self) -> List[DSCFile]:
        files: List[DSCFile] = []
        response = await self.service.send_receive_request({'XPCDictionary_sideChannel': uuid.uuid4(), 'DSCFilePaths': []})
        file_count = response['DSCFilePaths']
        for i in range(file_count):
            response = await self.service.receive_response()
            response = response['DSCFilePaths']
            file_transfer = response['fileTransfer']
            expected_length = file_transfer['expectedLength']
            file_path = response['filePath']
            files.append(DSCFile(file_path=file_path, file_size=expected_length))
        return files

    async def download(self, out: Path) -> None:
        files = await self.get_dsc_file_list()
        for i, file in enumerate(files):
            self.logger.info(f'Downloading {file}')
            out_file = out / file.file_path[1:]  # trim the "/" prefix
            out_file.parent.mkdir(parents=True, exist_ok=True)
            with open(out_file, 'wb') as f:
                with tqdm(total=files[i].file_size, dynamic_ncols=True) as pb:
                    async for chunk in self.service.iter_file_chunks(files[i].file_size, file_idx=i):
                        f.write(chunk)
                        pb.update(len(chunk))



import pytest

from pymobiledevice3.services.accessibilityaudit import AccessibilityAudit


@pytest.fixture(scope='function')
def accessibility_audit(lockdown):
    with AccessibilityAudit(lockdown=lockdown) as accessibility_audit:
        yield accessibility_audit


def test_capabilities(accessibility_audit):
    assert 'deviceApiVersion' in accessibility_audit.capabilities


def test_invert_colors_in_settings(accessibility_audit):
    found = False
    for setting in accessibility_audit.settings:
        if setting.key == 'INVERT_COLORS':
            found = True
            break
    assert found



import logging
import time

import click

from pymobiledevice3.cli.cli_common import Command, print_json
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.diagnostics import DiagnosticsService

logger = logging.getLogger(__name__)


@click.group()
def cli() -> None:
    pass


@cli.group()
def diagnostics() -> None:
    """ Reboot/Shutdown device or access other diagnostics services """
    pass


@diagnostics.command('restart', cls=Command)
def diagnostics_restart(service_provider: LockdownClient):
    """ Restart device """
    DiagnosticsService(lockdown=service_provider).restart()


@diagnostics.command('shutdown', cls=Command)
def diagnostics_shutdown(service_provider: LockdownClient):
    """ Shutdown device """
    DiagnosticsService(lockdown=service_provider).shutdown()


@diagnostics.command('sleep', cls=Command)
def diagnostics_sleep(service_provider: LockdownClient):
    """ Put device into sleep """
    DiagnosticsService(lockdown=service_provider).sleep()


@diagnostics.command('info', cls=Command)
def diagnostics_info(service_provider: LockdownClient):
    """ Get diagnostics info """
    print_json(DiagnosticsService(lockdown=service_provider).info())


@diagnostics.command('ioregistry', cls=Command)
@click.option('--plane')
@click.option('--name')
@click.option('--ioclass')
def diagnostics_ioregistry(service_provider: LockdownClient, plane, name, ioclass):
    """ Get ioregistry info """
    print_json(DiagnosticsService(lockdown=service_provider).ioregistry(plane=plane, name=name, ioclass=ioclass))


@diagnostics.command('mg', cls=Command)
@click.argument('keys', nargs=-1, default=None)
def diagnostics_mg(service_provider: LockdownClient, keys):
    """ Get MobileGestalt key values from given list. If empty, return all known. """
    print_json(DiagnosticsService(lockdown=service_provider).mobilegestalt(keys=keys))


@diagnostics.group('battery')
def diagnostics_battery():
    """ Battery options """
    pass


@diagnostics_battery.command('single', cls=Command)
def diagnostics_battery_single(service_provider: LockdownClient):
    """ get single snapshot of battery data """
    raw_info = DiagnosticsService(lockdown=service_provider).get_battery()
    print_json(raw_info)


@diagnostics_battery.command('monitor', cls=Command)
def diagnostics_battery_monitor(service_provider: LockdownClient):
    """ monitor battery usage """
    diagnostics = DiagnosticsService(lockdown=service_provider)
    while True:
        raw_info = diagnostics.get_battery()
        info = {
            'InstantAmperage': raw_info.get('InstantAmperage'),
            'Temperature': raw_info.get('Temperature'),
            'Voltage': raw_info.get('Voltage'),
            'IsCharging': raw_info.get('IsCharging'),
            'CurrentCapacity': raw_info.get('CurrentCapacity'),
        }
        logger.info(info)
        time.sleep(1)


@diagnostics.command('wifi', cls=Command)
def diagnostics_wifi(service_provider: LockdownServiceProvider) -> None:
    """ Query WiFi info from IORegistry """
    raw_info = DiagnosticsService(lockdown=service_provider).get_wifi()
    print_json(raw_info)



from typing import Mapping

from pymobiledevice3.exceptions import PyMobileDevice3Exception
from pymobiledevice3.remote.remote_service import RemoteService
from pymobiledevice3.remote.remote_service_discovery import RemoteServiceDiscoveryService


class RestoreService(RemoteService):
    SERVICE_NAME = 'com.apple.RestoreRemoteServices.restoreserviced'

    def __init__(self, lockdown: RemoteServiceDiscoveryService):
        super().__init__(lockdown, self.SERVICE_NAME)

    async def delay_recovery_image(self) -> None:
        """
        Set `delay-recovery-image` on devices of ProductType 0x1677b394. Otherwise, fail
        """
        await self.validate_command('delayrecoveryimage')

    async def enter_recovery(self) -> None:
        """ Enter recovery """
        await self.validate_command('recovery')

    async def reboot(self) -> None:
        """ Reboot device """
        await self.validate_command('reboot')

    async def get_preflightinfo(self) -> Mapping:
        """ Get preflight info """
        return await self.service.send_receive_request({'command': 'getpreflightinfo'})

    async def get_nonces(self) -> Mapping:
        """ Get ApNonce and SEPNonce """
        return await self.service.send_receive_request({'command': 'getnonces'})

    async def get_app_parameters(self) -> Mapping:
        return await self.validate_command('getappparameters')

    async def restore_lang(self, language: str) -> Mapping:
        return await self.service.send_receive_request({'command': 'restorelang', 'argument': language})

    async def validate_command(self, command: str) -> Mapping:
        """ Execute command and validate result is `success` """
        response = await self.service.send_receive_request({'command': command})
        if response.get('result') != 'success':
            raise PyMobileDevice3Exception(f'request command: {command} failed with error: {response}')
        return response



from pymobiledevice3.exceptions import PyMobileDevice3Exception
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.afc import AfcService, AfcShell

VEND_CONTAINER = 'VendContainer'
VEND_DOCUMENTS = 'VendDocuments'

DOCUMENTS_ROOT = '/Documents'


class HouseArrestService(AfcService):
    SERVICE_NAME = 'com.apple.mobile.house_arrest'
    RSD_SERVICE_NAME = 'com.apple.mobile.house_arrest.shim.remote'

    def __init__(self, lockdown: LockdownServiceProvider, bundle_id: str, documents_only: bool = False):
        if isinstance(lockdown, LockdownClient):
            super().__init__(lockdown, self.SERVICE_NAME)
        else:
            super().__init__(lockdown, self.RSD_SERVICE_NAME)
        if documents_only:
            cmd = VEND_DOCUMENTS
        else:
            cmd = VEND_CONTAINER
        self.documents_only = documents_only
        try:
            self.send_command(bundle_id, cmd)
        except PyMobileDevice3Exception:
            self.close()

    def send_command(self, bundle_id: str, cmd: str = 'VendContainer') -> None:
        response = self.service.send_recv_plist({'Command': cmd, 'Identifier': bundle_id})
        error = response.get('Error')
        if error:
            raise PyMobileDevice3Exception(error)

    def shell(self) -> None:
        AfcShell.create(self.lockdown, service=self, auto_cd=DOCUMENTS_ROOT if self.documents_only else '/')



from typing import List

import click

from pymobiledevice3.cli.cli_common import Command, print_json
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.house_arrest import HouseArrestService
from pymobiledevice3.services.installation_proxy import InstallationProxyService


@click.group()
def cli() -> None:
    pass


@cli.group()
def apps() -> None:
    """ Manage installed applications """
    pass


@apps.command('list', cls=Command)
@click.option('app_type', '-t', '--type', type=click.Choice(['System', 'User', 'Hidden', 'Any']), default='Any',
              help='include only applications of given type')
@click.option('--calculate-sizes/--no-calculate-size', default=False)
def apps_list(service_provider: LockdownServiceProvider, app_type: str, calculate_sizes: bool) -> None:
    """ list installed apps """
    print_json(InstallationProxyService(lockdown=service_provider).get_apps(application_type=app_type,
                                                                            calculate_sizes=calculate_sizes))


@apps.command('query', cls=Command)
@click.argument('bundle_identifiers', nargs=-1)
@click.option('--calculate-sizes/--no-calculate-size', default=False)
def apps_query(service_provider: LockdownServiceProvider, bundle_identifiers: List[str], calculate_sizes: bool) -> None:
    """ query installed apps """
    print_json(InstallationProxyService(lockdown=service_provider)
               .get_apps(calculate_sizes=calculate_sizes, bundle_identifiers=bundle_identifiers))


@apps.command('uninstall', cls=Command)
@click.argument('bundle_id')
def uninstall(service_provider: LockdownClient, bundle_id):
    """ uninstall app by given bundle_id """
    InstallationProxyService(lockdown=service_provider).uninstall(bundle_id)


@apps.command('install', cls=Command)
@click.argument('ipa_or_app_path', type=click.Path(exists=True))
def install(service_provider: LockdownServiceProvider, ipa_or_app_path: str) -> None:
    """ install given .ipa/.app """
    InstallationProxyService(lockdown=service_provider).install_from_local(ipa_or_app_path)


@apps.command('afc', cls=Command)
@click.option('--documents', is_flag=True)
@click.argument('bundle_id')
def afc(service_provider: LockdownClient, bundle_id: str, documents: bool):
    """ open an AFC shell for given bundle_id, assuming its profile is installed """
    HouseArrestService(lockdown=service_provider, bundle_id=bundle_id, documents_only=documents).shell()



import logging
import platform
import plistlib
import uuid
from contextlib import suppress
from pathlib import Path
from typing import Generator, Mapping, Optional

from pymobiledevice3 import usbmux
from pymobiledevice3.common import get_home_folder
from pymobiledevice3.exceptions import MuxException, NotPairedError
from pymobiledevice3.osu.os_utils import get_os_utils
from pymobiledevice3.usbmux import PlistMuxConnection

logger = logging.getLogger(__name__)
OSUTILS = get_os_utils()
PAIRING_RECORD_EXT = 'plist'


def generate_host_id(hostname: str = None) -> str:
    hostname = platform.node() if hostname is None else hostname
    host_id = uuid.uuid3(uuid.NAMESPACE_DNS, hostname)
    return str(host_id).upper()


def get_itunes_pairing_record(identifier: str) -> Optional[Mapping]:
    filename = OSUTILS.pair_record_path / f'{identifier}.plist'
    try:
        with open(filename, 'rb') as f:
            pair_record = plistlib.load(f)
    except (PermissionError, FileNotFoundError, plistlib.InvalidFileException):
        return None
    return pair_record


def get_local_pairing_record(identifier: str, pairing_records_cache_folder: Path) -> Optional[Mapping]:
    logger.debug('Looking for pymobiledevice3 pairing record')
    path = pairing_records_cache_folder / f'{identifier}.{PAIRING_RECORD_EXT}'
    if not path.exists():
        logger.debug(f'No pymobiledevice3 pairing record found for device {identifier}')
        return None
    return plistlib.loads(path.read_bytes())


def get_preferred_pair_record(identifier: str, pairing_records_cache_folder: Path,
                              usbmux_address: Optional[str] = None) -> Mapping:
    """
    look for an existing pair record to connected device by following order:
    - usbmuxd
    - iTunes
    - local storage
    """

    # usbmuxd
    with suppress(NotPairedError, MuxException):
        with usbmux.create_mux(usbmux_address=usbmux_address) as mux:
            if isinstance(mux, PlistMuxConnection):
                pair_record = mux.get_pair_record(identifier)
                if pair_record is not None:
                    return pair_record

    # iTunes
    pair_record = get_itunes_pairing_record(identifier)
    if pair_record is not None:
        return pair_record

    # local storage
    return get_local_pairing_record(identifier, pairing_records_cache_folder)


def create_pairing_records_cache_folder(pairing_records_cache_folder: Path = None) -> Path:
    if pairing_records_cache_folder is None:
        pairing_records_cache_folder = get_home_folder()
    else:
        pairing_records_cache_folder.mkdir(parents=True, exist_ok=True)
    OSUTILS.chown_to_non_sudo_if_needed(pairing_records_cache_folder)
    return pairing_records_cache_folder


def get_remote_pairing_record_filename(identifier: str) -> str:
    return f'remote_{identifier}'


def iter_remote_pair_records() -> Generator[Path, None, None]:
    return get_home_folder().glob('remote_*')


def iter_remote_paired_identifiers() -> Generator[str, None, None]:
    for file in iter_remote_pair_records():
        yield file.parts[-1].split('remote_', 1)[1].split('.', 1)[0]



#!/usr/bin/env python3

import logging
import os
import pathlib
import posixpath
import shlex
import shutil
import stat as stat_module
import struct
import sys
from collections import namedtuple
from datetime import datetime
from re import Pattern
from typing import Callable, List, Optional, Union

import hexdump
from click.exceptions import Exit
from construct import Const, Container, CString, Enum, GreedyRange, Int64ul, Struct, Tell
from parameter_decorators import path_to_str
from pygments import formatters, highlight, lexers
from pygnuutils.cli.ls import ls as ls_cli
from pygnuutils.ls import Ls, LsStub
from tqdm import trange
from xonsh.built_ins import XSH
from xonsh.cli_utils import Annotated, Arg, ArgParserAlias
from xonsh.main import main as xonsh_main
from xonsh.tools import print_color

from pymobiledevice3.exceptions import AfcException, AfcFileNotFoundError, ArgumentError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.lockdown_service import LockdownService
from pymobiledevice3.utils import try_decode

MAXIMUM_READ_SIZE = 4 * 1024 ** 2  # 4 MB
MODE_MASK = 0o0000777

StatResult = namedtuple('StatResult',
                        ['st_mode', 'st_ino', 'st_dev', 'st_nlink', 'st_uid', 'st_gid', 'st_size', 'st_atime',
                         'st_mtime', 'st_ctime', 'st_blocks', 'st_blksize', 'st_birthtime'])

afc_opcode_t = Enum(Int64ul,
                    STATUS=0x00000001,
                    DATA=0x00000002,  # Data */
                    READ_DIR=0x00000003,  # ReadDir */
                    READ_FILE=0x00000004,  # ReadFile */
                    WRITE_FILE=0x00000005,  # WriteFile */
                    WRITE_PART=0x00000006,  # WritePart */
                    TRUNCATE=0x00000007,  # TruncateFile */
                    REMOVE_PATH=0x00000008,  # RemovePath */
                    MAKE_DIR=0x00000009,  # MakeDir */
                    GET_FILE_INFO=0x0000000a,  # GetFileInfo */
                    GET_DEVINFO=0x0000000b,  # GetDeviceInfo */
                    WRITE_FILE_ATOM=0x0000000c,  # WriteFileAtomic (tmp file+rename) */
                    FILE_OPEN=0x0000000d,  # FileRefOpen */
                    FILE_OPEN_RES=0x0000000e,  # FileRefOpenResult */
                    READ=0x0000000f,  # FileRefRead */
                    WRITE=0x00000010,  # FileRefWrite */
                    FILE_SEEK=0x00000011,  # FileRefSeek */
                    FILE_TELL=0x00000012,  # FileRefTell */
                    FILE_TELL_RES=0x00000013,  # FileRefTellResult */
                    FILE_CLOSE=0x00000014,  # FileRefClose */
                    FILE_SET_SIZE=0x00000015,  # FileRefSetFileSize (ftruncate) */
                    GET_CON_INFO=0x00000016,  # GetConnectionInfo */
                    SET_CON_OPTIONS=0x00000017,  # SetConnectionOptions */
                    RENAME_PATH=0x00000018,  # RenamePath */
                    SET_FS_BS=0x00000019,  # SetFSBlockSize (0x800000) */
                    SET_SOCKET_BS=0x0000001A,  # SetSocketBlockSize (0x800000) */
                    FILE_LOCK=0x0000001B,  # FileRefLock */
                    MAKE_LINK=0x0000001C,  # MakeLink */
                    SET_FILE_TIME=0x0000001E,  # set st_mtime */
                    )

afc_error_t = Enum(Int64ul,
                   SUCCESS=0,
                   UNKNOWN_ERROR=1,
                   OP_HEADER_INVALID=2,
                   NO_RESOURCES=3,
                   READ_ERROR=4,
                   WRITE_ERROR=5,
                   UNKNOWN_PACKET_TYPE=6,
                   INVALID_ARG=7,
                   OBJECT_NOT_FOUND=8,
                   OBJECT_IS_DIR=9,
                   PERM_DENIED=10,
                   SERVICE_NOT_CONNECTED=11,
                   OP_TIMEOUT=12,
                   TOO_MUCH_DATA=13,
                   END_OF_DATA=14,
                   OP_NOT_SUPPORTED=15,
                   OBJECT_EXISTS=16,
                   OBJECT_BUSY=17,
                   NO_SPACE_LEFT=18,
                   OP_WOULD_BLOCK=19,
                   IO_ERROR=20,
                   OP_INTERRUPTED=21,
                   OP_IN_PROGRESS=22,
                   INTERNAL_ERROR=23,
                   MUX_ERROR=30,
                   NO_MEM=31,
                   NOT_ENOUGH_DATA=32,
                   DIR_NOT_EMPTY=33,
                   )

afc_link_type_t = Enum(Int64ul,
                       HARDLINK=1,
                       SYMLINK=2,
                       )

afc_fopen_mode_t = Enum(Int64ul,
                        RDONLY=0x00000001,  # /**< r   O_RDONLY */
                        RW=0x00000002,  # /**< r+  O_RDWR   | O_CREAT */
                        WRONLY=0x00000003,  # /**< w   O_WRONLY | O_CREAT  | O_TRUNC */
                        WR=0x00000004,  # /**< w+  O_RDWR   | O_CREAT  | O_TRUNC */
                        APPEND=0x00000005,  # /**< a   O_WRONLY | O_APPEND | O_CREAT */
                        RDAPPEND=0x00000006,  # /**< a+  O_RDWR   | O_APPEND | O_CREAT */
                        )

AFC_FOPEN_TEXTUAL_MODES = {
    'r': afc_fopen_mode_t.RDONLY,
    'r+': afc_fopen_mode_t.RW,
    'w': afc_fopen_mode_t.WRONLY,
    'w+': afc_fopen_mode_t.WR,
    'a': afc_fopen_mode_t.APPEND,
    'a+': afc_fopen_mode_t.RDAPPEND,
}

AFC_LOCK_SH = 1 | 4  # /**< shared lock */
AFC_LOCK_EX = 2 | 4  # /**< exclusive lock */
AFC_LOCK_UN = 8 | 4  # /**< unlock */

MAXIMUM_WRITE_SIZE = 1 << 30

AFCMAGIC = b'CFA6LPAA'

afc_header_t = Struct(
    'magic' / Const(AFCMAGIC),
    'entire_length' / Int64ul,
    'this_length' / Int64ul,
    'packet_num' / Int64ul,
    'operation' / afc_opcode_t,
    '_data_offset' / Tell,
)

afc_read_dir_req_t = Struct(
    'filename' / CString('utf8'),
)

afc_read_dir_resp_t = Struct(
    'filenames' / GreedyRange(CString('utf8')),
)

afc_mkdir_req_t = Struct(
    'filename' / CString('utf8'),
)

afc_stat_t = Struct(
    'filename' / CString('utf8'),
)

afc_make_link_req_t = Struct(
    'type' / afc_link_type_t,
    'target' / CString('utf8'),
    'source' / CString('utf8'),
)

afc_fopen_req_t = Struct(
    'mode' / afc_fopen_mode_t,
    'filename' / CString('utf8'),
)

afc_fopen_resp_t = Struct(
    'handle' / Int64ul,
)

afc_fclose_req_t = Struct(
    'handle' / Int64ul,
)

afc_rm_req_t = Struct(
    'filename' / CString('utf8'),
)

afc_rename_req_t = Struct(
    'source' / CString('utf8'),
    'target' / CString('utf8'),
)

afc_fread_req_t = Struct(
    'handle' / Int64ul,
    'size' / Int64ul,
)

afc_lock_t = Struct(
    'handle' / Int64ul,
    'op' / Int64ul,
)


def list_to_dict(d):
    d = d.decode('utf-8')
    t = d.split('\x00')
    t = t[:-1]

    assert len(t) % 2 == 0
    res = {}
    for i in range(int(len(t) / 2)):
        res[t[i * 2]] = t[i * 2 + 1]
    return res


class AfcService(LockdownService):
    SERVICE_NAME = 'com.apple.afc'
    RSD_SERVICE_NAME = 'com.apple.afc.shim.remote'

    def __init__(self, lockdown: LockdownServiceProvider, service_name: str = None):
        if service_name is None:
            if isinstance(lockdown, LockdownClient):
                service_name = self.SERVICE_NAME
            else:
                service_name = self.RSD_SERVICE_NAME
        super().__init__(lockdown, service_name)
        self.packet_num = 0

    def pull(self, relative_src: str, dst: str, match: Optional[Pattern] = None, callback: Optional[Callable] = None,
             src_dir: str = '') -> None:
        src = self.resolve_path(posixpath.join(src_dir, relative_src))

        if not self.isdir(src):
            # normal file
            if os.path.isdir(dst):
                dst = os.path.join(dst, os.path.basename(relative_src))
            with open(dst, 'wb') as f:
                src_size = self.stat(src)['st_size']
                if src_size <= MAXIMUM_READ_SIZE:
                    f.write(self.get_file_contents(src))
                else:
                    left_size = src_size
                    handle = self.fopen(src)
                    for _ in trange(src_size // MAXIMUM_READ_SIZE + 1):
                        f.write(self.fread(handle, min(MAXIMUM_READ_SIZE, left_size)))
                        left_size -= MAXIMUM_READ_SIZE
                    self.fclose(handle)
            os.utime(dst, (os.stat(dst).st_atime, self.stat(src)['st_mtime'].timestamp()))
            if callback is not None:
                callback(src, dst)
        else:
            # directory
            dst_path = pathlib.Path(dst) / os.path.basename(relative_src)
            dst_path.mkdir(parents=True, exist_ok=True)

            for filename in self.listdir(src):
                src_filename = posixpath.join(src, filename)
                dst_filename = dst_path / filename

                src_filename = self.resolve_path(src_filename)

                if match is not None and not match.match(posixpath.basename(src_filename)):
                    continue

                if self.isdir(src_filename):
                    dst_filename.mkdir(exist_ok=True)
                    self.pull(src_filename, str(dst_path), callback=callback)
                    continue

                self.pull(src_filename, str(dst_path), callback=callback)

    @path_to_str()
    def exists(self, filename):
        try:
            self.stat(filename)
            return True
        except AfcFileNotFoundError:
            return False

    @path_to_str()
    def wait_exists(self, filename):
        while not self.exists(filename):
            pass

    @path_to_str()
    def _push_internal(self, local_path, remote_path, callback=None):
        if callback is not None:
            callback(local_path, remote_path)

        if not os.path.isdir(local_path):
            # normal file
            try:
                if self.isdir(remote_path):
                    # Remote is dir.
                    remote_path = posixpath.join(remote_path, os.path.basename(local_path))
            except AfcFileNotFoundError:
                # Remote is file.
                remote_parent = posixpath.dirname(remote_path)
                if not self.exists(remote_parent):
                    raise
                remote_path = posixpath.join(remote_parent, os.path.basename(remote_path))
            with open(local_path, 'rb') as f:
                self.set_file_contents(remote_path, f.read())
        else:
            # directory
            if not self.exists(remote_path):
                self.makedirs(remote_path)

            for filename in os.listdir(local_path):
                local_filename = os.path.join(local_path, filename)
                remote_filename = posixpath.join(remote_path, filename).removeprefix('/')

                if os.path.isdir(local_filename):
                    if not self.exists(remote_filename):
                        self.makedirs(remote_filename)
                    self._push_internal(local_filename, remote_filename, callback=callback)
                    continue

                self._push_internal(local_filename, remote_filename, callback=callback)

    @path_to_str()
    def push(self, local_path, remote_path, callback=None):
        if os.path.isdir(local_path):
            remote_path = posixpath.join(remote_path, os.path.basename(local_path))
        self._push_internal(local_path, remote_path, callback)

    @path_to_str()
    def rm_single(self, filename: str, force: bool = False) -> bool:
        """ remove single file or directory

         return if succeed or raise exception depending on force parameter.

        :param filename: path to directory or a file
        :param force: True for ignore exception and return False
        :return: if succeed
        :rtype: bool
        """
        try:
            self._do_operation(afc_opcode_t.REMOVE_PATH, afc_rm_req_t.build({'filename': filename}))
            return True
        except AfcException:
            if force:
                return False
            raise

    @path_to_str()
    def rm(self, filename: str, match: Optional[Pattern] = None, force: bool = False) -> List[str]:
        """ recursive removal of a directory or a file

        if did not succeed, return list of undeleted filenames or raise exception depending on force parameter.

        :param filename: path to directory or a file
        :param match: Pattern of directory entries to remove or None to remove all
        :param force: True for ignore exception and return list of undeleted paths
        :return: list of undeleted paths
        :rtype: list[str]
        """
        if not self.exists(filename):
            if not self.rm_single(filename, force=force):
                return [filename]

        # single file
        if not self.isdir(filename):
            if self.rm_single(filename, force=force):
                return []
            return [filename]

        # directory content
        undeleted_items = []
        for entry in self.listdir(filename):
            current_filename = posixpath.join(filename, entry)

            if match is not None and not match.match(posixpath.basename(current_filename)):
                continue

            if self.isdir(current_filename):
                ret_undeleted_items = self.rm(current_filename, force=True)
                undeleted_items.extend(ret_undeleted_items)
            else:
                if not self.rm_single(current_filename, force=True):
                    undeleted_items.append(current_filename)

        # directory path
        try:
            if not self.rm_single(filename, force=force):
                undeleted_items.append(filename)
                return undeleted_items
        except AfcException:
            if undeleted_items:
                undeleted_items.append(filename)
            else:
                raise

        if undeleted_items:
            raise AfcException(f'Failed to delete paths: {undeleted_items}', None)

        return []

    def get_device_info(self):
        return list_to_dict(self._do_operation(afc_opcode_t.GET_DEVINFO))

    @path_to_str()
    def listdir(self, filename: str):
        data = self._do_operation(afc_opcode_t.READ_DIR, afc_read_dir_req_t.build({'filename': filename}))
        return afc_read_dir_resp_t.parse(data).filenames[2:]  # skip the . and ..

    @path_to_str()
    def makedirs(self, filename: str):
        return self._do_operation(afc_opcode_t.MAKE_DIR, afc_mkdir_req_t.build({'filename': filename}))

    @path_to_str()
    def isdir(self, filename: str) -> bool:
        stat = self.stat(filename)
        return stat.get('st_ifmt') == 'S_IFDIR'

    @path_to_str()
    def stat(self, filename: str):
        try:
            stat = list_to_dict(
                self._do_operation(afc_opcode_t.GET_FILE_INFO, afc_stat_t.build({'filename': filename})))
        except AfcException as e:
            if e.status != afc_error_t.READ_ERROR:
                raise
            raise AfcFileNotFoundError(e.args[0], e.status) from e

        stat['st_size'] = int(stat['st_size'])
        stat['st_blocks'] = int(stat['st_blocks'])
        stat['st_mtime'] = int(stat['st_mtime'])
        stat['st_birthtime'] = int(stat['st_birthtime'])
        stat['st_nlink'] = int(stat['st_nlink'])
        stat['st_mtime'] = datetime.fromtimestamp(stat['st_mtime'] / (10 ** 9))
        stat['st_birthtime'] = datetime.fromtimestamp(stat['st_birthtime'] / (10 ** 9))
        return stat

    @path_to_str()
    def os_stat(self, path: str):
        stat = self.stat(path)
        mode = 0
        for s_mode in ['S_IFDIR', 'S_IFCHR', 'S_IFBLK', 'S_IFREG', 'S_IFIFO', 'S_IFLNK', 'S_IFSOCK']:
            if stat['st_ifmt'] == s_mode:
                mode = getattr(stat_module, s_mode)
        return StatResult(
            mode, hash(posixpath.normpath(path)), 0, stat['st_nlink'], 0, 0, stat['st_size'],
            stat['st_mtime'].timestamp(), stat['st_mtime'].timestamp(), stat['st_birthtime'].timestamp(),
            stat['st_blocks'], 4096, stat['st_birthtime'].timestamp(),
        )

    @path_to_str()
    def link(self, target: str, source: str, type_=afc_link_type_t.SYMLINK):
        return self._do_operation(afc_opcode_t.MAKE_LINK,
                                  afc_make_link_req_t.build({'type': type_, 'target': target, 'source': source}))

    @path_to_str()
    def fopen(self, filename: str, mode: str = 'r') -> int:
        if mode not in AFC_FOPEN_TEXTUAL_MODES:
            raise ArgumentError(f'mode can be only one of: {AFC_FOPEN_TEXTUAL_MODES.keys()}')

        data = self._do_operation(afc_opcode_t.FILE_OPEN,
                                  afc_fopen_req_t.build({'mode': AFC_FOPEN_TEXTUAL_MODES[mode], 'filename': filename}))
        return afc_fopen_resp_t.parse(data).handle

    def fclose(self, handle: int):
        return self._do_operation(afc_opcode_t.FILE_CLOSE, afc_fclose_req_t.build({'handle': handle}))

    @path_to_str()
    def rename(self, source: str, target: str):
        try:
            return self._do_operation(afc_opcode_t.RENAME_PATH,
                                      afc_rename_req_t.build({'source': source, 'target': target}))
        except AfcException as e:
            if self.exists(source):
                raise
            raise AfcFileNotFoundError(e.args[0], e.status) from e

    def fread(self, handle: int, sz: bytes) -> bytes:
        data = b''
        while sz > 0:
            if sz > MAXIMUM_READ_SIZE:
                to_read = MAXIMUM_READ_SIZE
            else:
                to_read = sz
            self._dispatch_packet(afc_opcode_t.READ, afc_fread_req_t.build({'handle': handle, 'size': to_read}))
            status, chunk = self._receive_data()
            if status != afc_error_t.SUCCESS:
                raise AfcException('fread error', status)
            sz -= to_read
            data += chunk
        return data

    def fwrite(self, handle, data, chunk_size=MAXIMUM_WRITE_SIZE):
        file_handle = struct.pack('<Q', handle)
        chunks_count = len(data) // chunk_size
        b = b''
        for i in range(chunks_count):
            chunk = data[i * chunk_size:(i + 1) * chunk_size]
            self._dispatch_packet(afc_opcode_t.WRITE,
                                  file_handle + chunk,
                                  this_length=48)
            b += chunk

            status, response = self._receive_data()
            if status != afc_error_t.SUCCESS:
                raise AfcException(f'failed to write chunk: {status}', status)

        if len(data) % chunk_size:
            chunk = data[chunks_count * chunk_size:]
            self._dispatch_packet(afc_opcode_t.WRITE,
                                  file_handle + chunk,
                                  this_length=48)

            b += chunk

            status, response = self._receive_data()
            if status != afc_error_t.SUCCESS:
                raise AfcException(f'failed to write last chunk: {status}', status)

    @path_to_str()
    def resolve_path(self, filename: str):
        info = self.stat(filename)
        if info['st_ifmt'] == 'S_IFLNK':
            target = info['LinkTarget']
            if not target.startswith('/'):
                # relative path
                filename = posixpath.join(posixpath.dirname(filename), target)
            else:
                filename = target
        return filename

    @path_to_str()
    def get_file_contents(self, filename):
        filename = self.resolve_path(filename)
        info = self.stat(filename)

        if info['st_ifmt'] != 'S_IFREG':
            raise AfcException(f'{filename} isn\'t a file', afc_error_t.INVALID_ARG)

        h = self.fopen(filename)
        if not h:
            return
        d = self.fread(h, int(info['st_size']))
        self.fclose(h)
        return d

    @path_to_str()
    def set_file_contents(self, filename: str, data: bytes) -> None:
        h = self.fopen(filename, 'w')
        self.fwrite(h, data)
        self.fclose(h)

    @path_to_str()
    def walk(self, dirname: str):
        dirs = []
        files = []
        for fd in self.listdir(dirname):
            if fd in ('.', '..', ''):
                continue
            infos = self.stat(posixpath.join(dirname, fd))
            if infos and infos.get('st_ifmt') == 'S_IFDIR':
                dirs.append(fd)
            else:
                files.append(fd)

        yield dirname, dirs, files

        if dirs:
            for d in dirs:
                yield from self.walk(posixpath.join(dirname, d))

    @path_to_str()
    def dirlist(self, root, depth=-1):
        for folder, dirs, files in self.walk(root):
            if folder == root:
                yield folder
                if depth == 0:
                    break
            if folder != root and depth != -1 and folder.count(posixpath.sep) >= depth:
                continue
            for entry in dirs + files:
                yield posixpath.join(folder, entry)

    def lock(self, handle, operation):
        return self._do_operation(afc_opcode_t.FILE_LOCK, afc_lock_t.build({'handle': handle, 'op': operation}))

    def _dispatch_packet(self, operation, data, this_length=0):
        afcpack = Container(magic=AFCMAGIC,
                            entire_length=afc_header_t.sizeof() + len(data),
                            this_length=afc_header_t.sizeof() + len(data),
                            packet_num=self.packet_num,
                            operation=operation)
        if this_length:
            afcpack.this_length = this_length
        header = afc_header_t.build(afcpack)
        self.packet_num += 1
        self.service.sendall(header + data)

    def _receive_data(self):
        res = self.service.recvall(afc_header_t.sizeof())
        status = afc_error_t.SUCCESS
        data = ''
        if res:
            res = afc_header_t.parse(res)
            assert res['entire_length'] >= afc_header_t.sizeof()
            length = res['entire_length'] - afc_header_t.sizeof()
            data = self.service.recvall(length)
            if res.operation == afc_opcode_t.STATUS:
                if length != 8:
                    self.logger.error('Status length != 8')
                status = afc_error_t.parse(data)
            elif res.operation != afc_opcode_t.DATA:
                pass
        return status, data

    def _do_operation(self, opcode: afc_opcode_t, data: bytes = b''):
        self._dispatch_packet(opcode, data)
        status, data = self._receive_data()

        exception = AfcException
        if status != afc_error_t.SUCCESS:
            if status == afc_error_t.OBJECT_NOT_FOUND:
                exception = AfcFileNotFoundError

            raise exception(f'opcode: {opcode} failed with status: {status}', status)

        return data


class AfcLsStub(LsStub):
    def __init__(self, afc_shell, stdout):
        self.afc_shell = afc_shell
        self.stdout = stdout

    @property
    def sep(self):
        return posixpath.sep

    def join(self, path, *paths):
        return posixpath.join(path, *paths)

    def abspath(self, path):
        return posixpath.normpath(path)

    def stat(self, path, dir_fd=None, follow_symlinks=True):
        if follow_symlinks:
            path = self.afc_shell.afc.resolve_path(path)
        return self.afc_shell.afc.os_stat(path)

    def readlink(self, path, dir_fd=None):
        return self.afc_shell.afc.resolve_path(path)

    def isabs(self, path):
        return posixpath.isabs(path)

    def dirname(self, path):
        return posixpath.dirname(path)

    def basename(self, path):
        return posixpath.basename(path)

    def getgroup(self, st_gid):
        return '-'

    def getuser(self, st_uid):
        return '-'

    def now(self):
        return self.afc_shell.lockdown.date

    def listdir(self, path='.'):
        return self.afc_shell.afc.listdir(path)

    def system(self):
        return 'Darwin'

    def getenv(self, key, default=None):
        return ''

    def print(self, *objects, sep=' ', end='\n', file=sys.stdout, flush=False):
        print(objects[0], end=end)

    def get_tty_width(self):
        return os.get_terminal_size().columns


def path_completer(xsh, action, completer, alias, command):
    shell: AfcShell = XSH.ctx['_shell']
    pwd = shell.cwd
    is_absolute = command.prefix.startswith('/')
    dirpath = posixpath.join(pwd, command.prefix)
    if not shell.afc.exists(dirpath):
        dirpath = posixpath.dirname(dirpath)
    result = []
    for f in shell.afc.listdir(dirpath):
        if is_absolute:
            completion_option = posixpath.join(dirpath, f)
        else:
            completion_option = posixpath.relpath(posixpath.join(dirpath, f), pwd)
        try:
            if shell.afc.isdir(posixpath.join(dirpath, f)):
                result.append(f'{completion_option}/')
            else:
                result.append(completion_option)
        except AfcException:
            result.append(completion_option)
    return result


def dir_completer(xsh, action, completer, alias, command):
    shell: AfcShell = XSH.ctx['_shell']
    pwd = shell.cwd
    is_absolute = command.prefix.startswith('/')
    dirpath = posixpath.join(pwd, command.prefix)
    if not shell.afc.exists(dirpath):
        dirpath = posixpath.dirname(dirpath)
    result = []
    for f in shell.afc.listdir(dirpath):
        if is_absolute:
            completion_option = posixpath.join(dirpath, f)
        else:
            completion_option = posixpath.relpath(posixpath.join(dirpath, f), pwd)
        try:
            if shell.afc.isdir(posixpath.join(dirpath, f)):
                result.append(f'{completion_option}/')
        except AfcException:
            result.append(completion_option)
    return result


class AfcShell:
    @classmethod
    def create(cls, service_provider: LockdownServiceProvider, service_name: Optional[str] = None,
               service: Optional[LockdownService] = None, auto_cd: Optional[str] = '/'):
        args = ['--rc', str(pathlib.Path(__file__).absolute())]
        os.environ['XONSH_COLOR_STYLE'] = 'default'
        XSH.ctx['_class'] = cls
        XSH.ctx['_lockdown'] = service_provider
        XSH.ctx['_auto_cd'] = auto_cd
        if service is not None:
            XSH.ctx['_service'] = service
        else:
            XSH.ctx['_service'] = AfcService(service_provider, service_name=service_name)

        try:
            logging.getLogger('parso.python.diff').disabled = True
            logging.getLogger('parso.cache').disabled = True
            xonsh_main(args)
        except SystemExit:
            pass

    def __init__(self, lockdown: LockdownServiceProvider, service: AfcService):
        self.lockdown = lockdown
        self.afc = service
        XSH.ctx['_shell'] = self
        self.cwd = XSH.ctx.get('_auto_cd', '/')
        self._commands = {}
        self._orig_aliases = {}
        self._orig_prompt = XSH.env['PROMPT']
        self._setup_shell_commands()

        print_color('''
        {BOLD_WHITE}Welcome to xonsh-afc shell! 👋{RESET}
        Use {CYAN}show-help{RESET} to view a list of all available special commands.
            These special commands will replace all already existing commands.
        ''')

    def _register_arg_parse_alias(self, name: str, handler: Union[Callable, str]):
        handler = ArgParserAlias(func=handler, has_args=True, prog=name)
        self._commands[name] = handler
        if XSH.aliases.get(name):
            self._orig_aliases[name] = XSH.aliases[name]
        XSH.aliases[name] = handler

    def _register_rpc_command(self, name, handler):
        self._commands[name] = handler
        if XSH.aliases.get(name):
            self._orig_aliases[name] = XSH.aliases[name]
        XSH.aliases[name] = handler

    def _setup_shell_commands(self):
        # clear all host commands except for some useful ones
        XSH.env['PATH'].clear()
        # adding "file" just to fix xonsh errors
        for cmd in ['wc', 'grep', 'egrep', 'sed', 'awk', 'print', 'yes', 'cat', 'file']:
            executable = shutil.which(cmd)
            if executable is not None:
                self._register_rpc_command(cmd, executable)

        self._register_rpc_command('ls', self.do_ls)
        self._register_arg_parse_alias('pwd', self._do_pwd)
        self._register_arg_parse_alias('link', self._do_link)
        self._register_arg_parse_alias('cd', self._do_cd)
        self._register_arg_parse_alias('cat', self._do_cat)
        self._register_arg_parse_alias('rm', self._do_rm)
        self._register_arg_parse_alias('pull', self._do_pull)
        self._register_arg_parse_alias('push', self._do_push)
        self._register_arg_parse_alias('walk', self._do_walk)
        self._register_arg_parse_alias('head', self._do_head)
        self._register_arg_parse_alias('hexdump', self._do_hexdump)
        self._register_arg_parse_alias('mkdir', self._do_mkdir)
        self._register_arg_parse_alias('info', self._do_info)
        self._register_arg_parse_alias('mv', self._do_mv)
        self._register_arg_parse_alias('stat', self._do_stat)
        self._register_arg_parse_alias('show-help', self._do_show_help)

        XSH.env['PROMPT'] = f'[{{BOLD_CYAN}}{self.afc.service_name}:{{afc_cwd}}{{RESET}}]{{prompt_end}} '
        XSH.env['PROMPT_FIELDS']['afc_cwd'] = self._afc_cwd
        XSH.env['PROMPT_FIELDS']['prompt_end'] = self._prompt

    def _prompt(self) -> str:
        if len(XSH.history) == 0 or XSH.history[-1].rtn == 0:
            return '{BOLD_GREEN}${RESET}'
        return '{BOLD_RED}${RESET}'

    def _afc_cwd(self) -> str:
        return self.cwd

    def _relative_path(self, filename: str) -> str:
        return posixpath.join(self.cwd, filename)

    def _do_show_help(self):
        """
        list all rpc commands
        """
        buf = ''
        for k, v in self._commands.items():
            buf += f'👾 {k}\n'
        print(buf)

    def _do_pwd(self) -> None:
        print(self.cwd)

    def _do_link(self, target: str, source: str) -> None:
        self.afc.link(self.relative_path(target), self.relative_path(source), afc_link_type_t.SYMLINK)

    def _do_cd(self, directory: Annotated[str, Arg(completer=dir_completer)]) -> None:
        directory = self.relative_path(directory)
        directory = posixpath.normpath(directory)
        if self.afc.exists(directory):
            self.cwd = directory
            self._update_prompt()
        else:
            print(f'[ERROR] {directory} does not exist')

    def do_ls(self, args, stdin, stdout, stderr):
        """ list files """
        try:
            with ls_cli.make_context('ls', args) as ctx:
                files = list(map(self._relative_path, ctx.params.pop('files')))
                files = files if files else [self.cwd]
                Ls(AfcLsStub(self, stdout))(*files, **ctx.params)
        except Exit:
            pass

    def _do_walk(self, directory: Annotated[str, Arg(completer=dir_completer)]):
        for root, dirs, files in self.afc.walk(self.relative_path(directory)):
            for name in files:
                print(posixpath.join(root, name))
            for name in dirs:
                print(posixpath.join(root, name))

    def _do_cat(self, filename: str):
        print(try_decode(self.afc.get_file_contents(self.relative_path(filename))))

    def _do_rm(self, file: Annotated[List[str], Arg(nargs='+', completer=path_completer)]):
        for filename in file:
            self.afc.rm(self.relative_path(filename))

    def _do_pull(self, remote_path: Annotated[str, Arg(completer=path_completer)], local_path: str):
        def log(src, dst):
            print(f'{src} --> {dst}')

        self.afc.pull(remote_path, local_path, callback=log, src_dir=self.cwd)

    def _do_push(self, local_path: str, remote_path: Annotated[str, Arg(completer=path_completer)]):
        def log(src, dst):
            print(f'{src} --> {dst}')

        self.afc.push(local_path, self.relative_path(remote_path), callback=log)

    def _do_head(self, filename: Annotated[str, Arg(completer=path_completer)]):
        print(try_decode(self.afc.get_file_contents(self.relative_path(filename))[:32]))

    def _do_hexdump(self, filename: Annotated[str, Arg(completer=path_completer)]):
        print(hexdump.hexdump(self.afc.get_file_contents(self.relative_path(filename)), result='return'))

    def _do_mkdir(self, filename: Annotated[str, Arg(completer=path_completer)]):
        self.afc.makedirs(self.relative_path(filename))

    def _do_info(self):
        for k, v in self.afc.get_device_info().items():
            print(f'{k}: {v}')

    def _do_mv(self, source: Annotated[str, Arg(completer=path_completer)],
               dest: Annotated[str, Arg(completer=path_completer)]):
        return self.afc.rename(self.relative_path(source), self.relative_path(dest))

    def _do_stat(self, filename: Annotated[str, Arg(completer=path_completer)]):
        for k, v in self.afc.stat(self.relative_path(filename)).items():
            print(f'{k}: {v}')

    def relative_path(self, filename: str) -> str:
        return posixpath.join(self.cwd, filename)

    def _update_prompt(self) -> None:
        self.prompt = highlight(f'[{self.afc.service_name}:{self.cwd}]$ ', lexers.BashSessionLexer(),
                                formatters.TerminalTrueColorFormatter(style='solarized-dark')).strip()

    def _complete(self, text, line, begidx, endidx):
        curdir_diff = posixpath.dirname(text)
        dirname = posixpath.join(self.cwd, curdir_diff)
        prefix = posixpath.basename(text)
        return [
            str(posixpath.join(curdir_diff, filename)) for filename in self.afc.listdir(dirname)
            if filename.startswith(prefix)
        ]

    def _complete_first_arg(self, text, line, begidx, endidx):
        if self._count_completion_parts(line, begidx) > 1:
            return []
        return self._complete(text, line, begidx, endidx)

    def _complete_push_arg(self, text, line, begidx, endidx):
        count = self._count_completion_parts(line, begidx)
        if count == 1:
            return self._complete_local(text)
        elif count == 2:
            return self._complete(text, line, begidx, endidx)
        else:
            return []

    def _complete_pull_arg(self, text, line, begidx, endidx):
        count = self._count_completion_parts(line, begidx)
        if count == 1:
            return self._complete(text, line, begidx, endidx)
        elif count == 2:
            return self._complete_local(text)
        else:
            return []

    @staticmethod
    def _complete_local(text: str):
        path = pathlib.Path(text)
        path_iter = path.iterdir() if text.endswith(os.path.sep) else path.parent.iterdir()
        return [str(p) for p in path_iter if str(p).startswith(text)]

    @staticmethod
    def _count_completion_parts(line, begidx):
        # Strip the " for paths including spaces.
        return len(shlex.split(line[:begidx].rstrip('"')))


if __name__ == str(pathlib.Path(__file__).absolute()):
    rc = XSH.ctx['_class'](XSH.ctx['_lockdown'], XSH.ctx['_service'])
    # fix fzf conflicts
    XSH.env['fzf_history_binding'] = ""  # Ctrl+R
    XSH.env['fzf_ssh_binding'] = ""  # Ctrl+S
    XSH.env['fzf_file_binding'] = ""  # Ctrl+T
    XSH.env['fzf_dir_binding'] = ""  # Ctrl+G



from datetime import datetime, timedelta
from pathlib import Path
from typing import Optional

from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
from cryptography.hazmat.primitives.serialization import Encoding, NoEncryption, PrivateFormat, load_pem_public_key
from cryptography.x509 import Certificate
from cryptography.x509.oid import NameOID


def make_cert(key: RSAPrivateKey, public_key: RSAPublicKey, common_name: Optional[str] = None) -> Certificate:
    attributes = [x509.NameAttribute(NameOID.COMMON_NAME, common_name)] if common_name else []
    subject = issuer = x509.Name(attributes)
    cert = x509.CertificateBuilder()
    cert = cert.subject_name(subject)
    cert = cert.issuer_name(issuer)
    cert = cert.public_key(public_key)
    cert = cert.serial_number(1)
    now = datetime.now()
    now = now.replace(tzinfo=None)
    cert = cert.not_valid_before(now - timedelta(minutes=1))
    cert = cert.not_valid_after(now + timedelta(days=365 * 10))
    cert = cert.sign(key, hashes.SHA256())
    return cert


def dump_cert(cert):
    return cert.public_bytes(Encoding.PEM)


def ca_do_everything(device_public_key, private_key: Optional[rsa.RSAPrivateKey] = None):
    if private_key is None:
        private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)

    cert = make_cert(private_key, private_key.public_key())
    dev_key = load_pem_public_key(device_public_key)
    dev_cert = make_cert(private_key, dev_key, 'Device')
    return dump_cert(cert), private_key.private_bytes(Encoding.PEM, PrivateFormat.PKCS8, NoEncryption()), dump_cert(
        dev_cert)


def create_keybag_file(file: Path, common_name: str) -> None:
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
    cer = make_cert(private_key, private_key.public_key(), common_name)
    file.write_bytes(
        private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption()
        ) + cer.public_bytes(encoding=serialization.Encoding.PEM))



import asyncio
import dataclasses
import logging
import sys
import tempfile
from functools import partial
from typing import List, Mapping, Optional, TextIO

import click

from pymobiledevice3.bonjour import DEFAULT_BONJOUR_TIMEOUT, browse_remotepairing_manual_pairing
from pymobiledevice3.cli.cli_common import BaseCommand, RSDCommand, print_json, prompt_device_list, sudo_required, \
    user_requested_colored_output
from pymobiledevice3.common import get_home_folder
from pymobiledevice3.exceptions import NoDeviceConnectedError
from pymobiledevice3.pair_records import PAIRING_RECORD_EXT, get_remote_pairing_record_filename
from pymobiledevice3.remote.common import ConnectionType, TunnelProtocol
from pymobiledevice3.remote.module_imports import MAX_IDLE_TIMEOUT, start_tunnel, verify_tunnel_imports
from pymobiledevice3.remote.remote_service_discovery import RSD_PORT, RemoteServiceDiscoveryService
from pymobiledevice3.remote.tunnel_service import RemotePairingManualPairingService, get_core_device_tunnel_services, \
    get_remote_pairing_tunnel_services
from pymobiledevice3.remote.utils import get_rsds
from pymobiledevice3.tunneld import TUNNELD_DEFAULT_ADDRESS, TunneldRunner

logger = logging.getLogger(__name__)


async def browse_rsd(timeout: float = DEFAULT_BONJOUR_TIMEOUT) -> List[Mapping]:
    devices = []
    for rsd in await get_rsds(timeout):
        devices.append({'address': rsd.service.address[0],
                        'port': RSD_PORT,
                        'UniqueDeviceID': rsd.peer_info['Properties']['UniqueDeviceID'],
                        'ProductType': rsd.peer_info['Properties']['ProductType'],
                        'OSVersion': rsd.peer_info['Properties']['OSVersion']})
    return devices


async def browse_remotepairing(timeout: float = DEFAULT_BONJOUR_TIMEOUT) -> List[Mapping]:
    devices = []
    for remotepairing in await get_remote_pairing_tunnel_services(timeout):
        devices.append({'address': remotepairing.hostname,
                        'port': remotepairing.port,
                        'identifier': remotepairing.remote_identifier})
    return devices


async def cli_browse(timeout: float = DEFAULT_BONJOUR_TIMEOUT) -> None:
    print_json({
        'usb': await browse_rsd(timeout),
        'wifi': await browse_remotepairing(timeout),
    })


@click.group()
def cli() -> None:
    pass


@cli.group('remote')
def remote_cli() -> None:
    """ Create RemoteXPC tunnels """
    pass


@remote_cli.command('tunneld', cls=BaseCommand)
@click.option('--host', default=TUNNELD_DEFAULT_ADDRESS[0])
@click.option('--port', type=click.INT, default=TUNNELD_DEFAULT_ADDRESS[1])
@click.option('-d', '--daemonize', is_flag=True)
@click.option('-p', '--protocol', type=click.Choice([e.value for e in TunnelProtocol]),
              default=TunnelProtocol.QUIC.value)
@click.option('--usb/--no-usb', default=True, help='Enable usb monitoring')
@click.option('--wifi/--no-wifi', default=True, help='Enable wifi monitoring')
@click.option('--usbmux/--no-usbmux', default=True, help='Enable usbmux monitoring')
@click.option('--mobdev2/--no-mobdev2', default=True, help='Enable mobdev2 monitoring')
@sudo_required
def cli_tunneld(
        host: str, port: int, daemonize: bool, protocol: str, usb: bool, wifi: bool, usbmux: bool,
        mobdev2: bool) -> None:
    """ Start Tunneld service for remote tunneling """
    if not verify_tunnel_imports():
        return
    protocol = TunnelProtocol(protocol)
    tunneld_runner = partial(TunneldRunner.create, host, port, protocol=protocol, usb_monitor=usb, wifi_monitor=wifi,
                             usbmux_monitor=usbmux, mobdev2_monitor=mobdev2)
    if daemonize:
        try:
            from daemonize import Daemonize
        except ImportError:
            raise NotImplementedError('daemonizing is only supported on unix platforms')
        with tempfile.NamedTemporaryFile('wt') as pid_file:
            daemon = Daemonize(app=f'Tunneld {host}:{port}', pid=pid_file.name,
                               action=tunneld_runner)
            logger.info(f'starting Tunneld {host}:{port}')
            daemon.start()
    else:
        tunneld_runner()


@remote_cli.command('browse', cls=BaseCommand)
@click.option('--timeout', type=click.FLOAT, default=DEFAULT_BONJOUR_TIMEOUT, help='Bonjour timeout (in seconds)')
def browse(timeout: float) -> None:
    """ browse RemoteXPC devices using bonjour """
    asyncio.run(cli_browse(timeout), debug=True)


@remote_cli.command('rsd-info', cls=RSDCommand)
def rsd_info(service_provider: RemoteServiceDiscoveryService):
    """ show info extracted from RSD peer """
    print_json(service_provider.peer_info)


async def tunnel_task(
        service, secrets: Optional[TextIO] = None, script_mode: bool = False,
        max_idle_timeout: float = MAX_IDLE_TIMEOUT, protocol: TunnelProtocol = TunnelProtocol.QUIC) -> None:
    async with start_tunnel(
            service, secrets=secrets, max_idle_timeout=max_idle_timeout, protocol=protocol) as tunnel_result:
        logger.info('tunnel created')
        if script_mode:
            print(f'{tunnel_result.address} {tunnel_result.port}')
        else:
            if user_requested_colored_output():
                if secrets is not None:
                    print(click.style('Secrets: ', bold=True, fg='magenta') +
                          click.style(secrets.name, bold=True, fg='white'))
                print(click.style('Identifier: ', bold=True, fg='yellow') +
                      click.style(service.remote_identifier, bold=True, fg='white'))
                print(click.style('Interface: ', bold=True, fg='yellow') +
                      click.style(tunnel_result.interface, bold=True, fg='white'))
                print(click.style('Protocol: ', bold=True, fg='yellow') +
                      click.style(tunnel_result.protocol, bold=True, fg='white'))
                print(click.style('RSD Address: ', bold=True, fg='yellow') +
                      click.style(tunnel_result.address, bold=True, fg='white'))
                print(click.style('RSD Port: ', bold=True, fg='yellow') +
                      click.style(tunnel_result.port, bold=True, fg='white'))
                print(click.style('Use the follow connection option:\n', bold=True, fg='yellow') +
                      click.style(f'--rsd {tunnel_result.address} {tunnel_result.port}', bold=True, fg='cyan'))
            else:
                if secrets is not None:
                    print(f'Secrets: {secrets.name}')
                print(f'Identifier: {service.remote_identifier}')
                print(f'Interface: {tunnel_result.interface}')
                print(f'Protocol: {tunnel_result.protocol}')
                print(f'RSD Address: {tunnel_result.address}')
                print(f'RSD Port: {tunnel_result.port}')
                print(f'Use the follow connection option:\n'
                      f'--rsd {tunnel_result.address} {tunnel_result.port}')
        sys.stdout.flush()
        await tunnel_result.client.wait_closed()
        logger.info('tunnel was closed')


async def start_tunnel_task(
        connection_type: ConnectionType, secrets: TextIO, udid: Optional[str] = None, script_mode: bool = False,
        max_idle_timeout: float = MAX_IDLE_TIMEOUT, protocol: TunnelProtocol = TunnelProtocol.QUIC) -> None:
    if start_tunnel is None:
        raise NotImplementedError('failed to start the tunnel on your platform')
    get_tunnel_services = {
        connection_type.USB: get_core_device_tunnel_services,
        connection_type.WIFI: get_remote_pairing_tunnel_services,
    }
    tunnel_services = await get_tunnel_services[connection_type](udid=udid)
    if not tunnel_services:
        # no devices were found
        raise NoDeviceConnectedError()
    if len(tunnel_services) == 1 or udid is not None:
        # only one device found
        service = tunnel_services[0]
    else:
        # several devices were found, show prompt if none explicitly selected
        service = prompt_device_list(tunnel_services)

    await tunnel_task(service, secrets=secrets, script_mode=script_mode, max_idle_timeout=max_idle_timeout,
                      protocol=protocol)


@remote_cli.command('start-tunnel', cls=BaseCommand)
@click.option('-t', '--connection-type', type=click.Choice([e.value for e in ConnectionType], case_sensitive=False),
              default=ConnectionType.USB.value)
@click.option('--udid', help='UDID for a specific device to look for')
@click.option('--secrets', type=click.File('wt'), help='TLS keyfile for decrypting with Wireshark')
@click.option('--script-mode', is_flag=True,
              help='Show only HOST and port number to allow easy parsing from external shell scripts')
@click.option('--max-idle-timeout', type=click.FLOAT, default=MAX_IDLE_TIMEOUT,
              help='Maximum QUIC idle time (ping interval)')
@click.option('-p', '--protocol',
              type=click.Choice([e.value for e in TunnelProtocol], case_sensitive=False),
              default=TunnelProtocol.QUIC.value)
@sudo_required
def cli_start_tunnel(
        connection_type: ConnectionType, udid: Optional[str], secrets: TextIO, script_mode: bool,
        max_idle_timeout: float, protocol: str) -> None:
    """ start tunnel """
    if not verify_tunnel_imports():
        return
    asyncio.run(
        start_tunnel_task(
            ConnectionType(connection_type), secrets, udid, script_mode, max_idle_timeout=max_idle_timeout,
            protocol=TunnelProtocol(protocol)), debug=True)


@dataclasses.dataclass
class RemotePairingManualPairingDevice:
    ip: str
    port: int
    device_name: str
    identifier: str


async def start_remote_pair_task(device_name: str) -> None:
    if start_tunnel is None:
        raise NotImplementedError('failed to start the tunnel on your platform')

    devices: List[RemotePairingManualPairingDevice] = []
    for answer in await browse_remotepairing_manual_pairing():
        current_device_name = answer.properties[b'name'].decode()

        if device_name is not None and current_device_name != device_name:
            continue

        for ip in answer.ips:
            devices.append(RemotePairingManualPairingDevice(ip=ip, port=answer.port, device_name=current_device_name,
                                                            identifier=answer.properties[b'identifier'].decode()))

    if len(devices) > 0:
        device = prompt_device_list(devices)
    else:
        logger.error('No devices were found during bonjour browse')
        return

    async with RemotePairingManualPairingService(device.identifier, device.ip, device.port) as service:
        await service.connect(autopair=True)


@remote_cli.command('pair', cls=BaseCommand)
@click.option('--name', help='Device name for a specific device to look for')
def cli_pair(name: Optional[str]) -> None:
    """ start remote pairing for devices which allow """
    asyncio.run(start_remote_pair_task(name), debug=True)


@remote_cli.command('delete-pair', cls=BaseCommand)
@click.argument('udid')
@sudo_required
def cli_delete_pair(udid: str):
    """ delete a pairing record """
    pair_record_path = get_home_folder() / f'{get_remote_pairing_record_filename(udid)}.{PAIRING_RECORD_EXT}'
    pair_record_path.unlink()


async def cli_service_task(service_provider: RemoteServiceDiscoveryService, service_name: str) -> None:
    async with service_provider.start_remote_service(service_name) as service:
        service.shell()


@remote_cli.command('service', cls=RSDCommand)
@click.argument('service_name')
def cli_service(service_provider: RemoteServiceDiscoveryService, service_name: str) -> None:
    """ start an ipython shell for interacting with given service """
    asyncio.run(cli_service_task(service_provider, service_name), debug=True)



import asyncio
import struct
import uuid
from enum import IntEnum
from typing import AsyncGenerator, List, Mapping, Optional

from pymobiledevice3.exceptions import CoreDeviceError
from pymobiledevice3.remote.core_device.core_device_service import CoreDeviceService
from pymobiledevice3.remote.remote_service_discovery import RemoteServiceDiscoveryService
from pymobiledevice3.remote.xpc_message import XpcUInt64Type


class Domain(IntEnum):
    APP_DATA_CONTAINER = 1
    APP_GROUP_DATA_CONTAINER = 2
    TEMPORARY = 3
    SYSTEM_CRASH_LOGS = 5


APPLE_DOMAIN_DICT = {
    'appDataContainer': Domain.APP_DATA_CONTAINER,
    'appGroupDataContainer': Domain.APP_GROUP_DATA_CONTAINER,
    'temporary': Domain.TEMPORARY,
    'systemCrashLogs': Domain.SYSTEM_CRASH_LOGS
}


class FileServiceService(CoreDeviceService):
    """
    Filesystem control
    """

    CTRL_SERVICE_NAME = 'com.apple.coredevice.fileservice.control'

    def __init__(self, rsd: RemoteServiceDiscoveryService, domain: Domain) -> None:
        super().__init__(rsd, self.CTRL_SERVICE_NAME)
        self.domain: Domain = domain
        self.session: Optional[str] = None

    async def connect(self) -> None:
        await super().connect()
        response = await self.send_receive_request({
            'Cmd': 'CreateSession', 'Domain': XpcUInt64Type(self.domain), 'Identifier': '', 'Session': '',
            'User': 'mobile'})
        self.session = response['NewSessionID']

    async def retrieve_directory_list(self, path: str = '.') -> AsyncGenerator[List[str], None]:
        return (await self.send_receive_request({
            'Cmd': 'RetrieveDirectoryList', 'MessageUUID': str(uuid.uuid4()), 'Path': path, 'SessionID': self.session}
        ))['FileList']

    async def retrieve_file(self, path: str = '.') -> bytes:
        response = await self.send_receive_request({
            'Cmd': 'RetrieveFile', 'Path': path, 'SessionID': self.session}
        )
        data_service = self.rsd.get_service_port('com.apple.coredevice.fileservice.data')
        reader, writer = await asyncio.open_connection(self.service.address[0], data_service)
        writer.write(b'rwb!FILE' + struct.pack('>QQQQ', response['Response'], 0, response['NewFileID'], 0))
        await writer.drain()
        await reader.readexactly(0x24)
        return await reader.readexactly(struct.unpack('>I', await reader.readexactly(4))[0])

    async def send_receive_request(self, request: Mapping) -> Mapping:
        response = await self.service.send_receive_request(request)
        encoded_error = response.get('EncodedError')
        if encoded_error is not None:
            localized_description = response.get('LocalizedDescription')
            if localized_description is not None:
                raise CoreDeviceError(localized_description)
            raise CoreDeviceError(encoded_error)
        return response
